@@ -148,7 +148,7 @@ fn pulley_get_operands(inst: &mut Inst, collector: &mut impl OperandVisitor) {
             collector.reg_def(dst);
         }
 
-        Inst::Call { info } => {
+        Inst::Call { info } | Inst::PatchableCall { info } => {
             let CallInfo {
                 uses,
                 defs,
@@ -435,7 +435,8 @@ where
             }
             | Inst::Call { .. }
             | Inst::IndirectCall { .. }
-            | Inst::IndirectCallHost { .. } => true,
+            | Inst::IndirectCallHost { .. }
+            | Inst::PatchableCall { .. } => true,
             _ => false,
         }
     }
@@ -498,9 +499,10 @@ where
 
     fn call_type(&self) -> CallType {
         match &self.inst {
-            Inst::Call { .. } | Inst::IndirectCall { .. } | Inst::IndirectCallHost { .. } => {
-                CallType::Regular
-            }
+            Inst::Call { .. }
+            | Inst::IndirectCall { .. }
+            | Inst::IndirectCallHost { .. }
+            | Inst::PatchableCall { .. } => CallType::Regular,
 
             Inst::ReturnCall { .. } | Inst::ReturnIndirectCall { .. } => CallType::TailCall,
 
@@ -533,6 +535,16 @@ where
         todo!()
     }
 
+    fn gen_nop_unit() -> SmallVec<[u8; 8]> {
+        let mut bytes = smallvec::smallvec![];
+        let nop = pulley_interpreter::op::Nop {};
+        nop.encode(&mut bytes);
+        // NOP needs to be a 1-byte opcode so it can be used to
+        // overwrite a callsite of any length.
+        assert_eq!(bytes.len(), 1);
+        bytes
+    }
+
     fn rc_for_type(ty: Type) -> CodegenResult<(&'static [RegClass], &'static [Type])> {
         match ty {
             I8 => Ok((&[RegClass::Int], &[I8])),
@@ -712,6 +724,10 @@ impl Inst {
                 format!("indirect_call {callee}, {info:?}{try_call}")
             }
 
+            Inst::PatchableCall { info } => {
+                format!("patchable_call {info:?}")
+            }
+
             Inst::ReturnCall { info } => {
                 format!("return_call {info:?}")
             }
