@@ -1,6 +1,9 @@
 //! Defines `JITModule`.
 
-use crate::{compiled_blob::CompiledBlob, memory::BranchProtection, memory::Memory};
+use crate::{
+    compiled_blob::CompiledBlob,
+    memory::{BranchProtection, JITMemoryProvider, SystemMemoryProvider},
+};
 use cranelift_codegen::binemit::Reloc;
 use cranelift_codegen::isa::{OwnedTargetIsa, TargetIsa};
 use cranelift_codegen::settings::Configurable;
@@ -28,6 +31,7 @@ pub struct JITBuilder {
     symbols: HashMap<String, SendWrapper<*const u8>>,
     lookup_symbols: Vec<Box<dyn Fn(&str) -> Option<*const u8> + Send>>,
     libcall_names: Box<dyn Fn(ir::LibCall) -> String + Send + Sync>,
+    memory: Option<Box<dyn JITMemoryProvider + Send>>,
 }
 
 impl JITBuilder {
@@ -91,6 +95,7 @@ impl JITBuilder {
             symbols,
             lookup_symbols,
             libcall_names,
+            memory: None,
         }
     }
 
@@ -141,6 +146,14 @@ impl JITBuilder {
         self.lookup_symbols.push(symbol_lookup_fn);
         self
     }
+
+    /// Set the memory provider for the module.
+    ///
+    /// If unset, defaults to [`SystemMemoryProvider`].
+    pub fn memory_provider(&mut self, provider: Box<dyn JITMemoryProvider + Send>) -> &mut Self {
+        self.memory = Some(provider);
+        self
+    }
 }
 
 /// A wrapper that impls Send for the contents.
@@ -159,21 +172,15 @@ pub struct JITModule {
     symbols: RefCell<HashMap<String, SendWrapper<*const u8>>>,
     lookup_symbols: Vec<Box<dyn Fn(&str) -> Option<*const u8> + Send>>,
     libcall_names: Box<dyn Fn(ir::LibCall) -> String + Send + Sync>,
-    memory: MemoryHandle,
+    memory: Box<dyn JITMemoryProvider + Send>,
     declarations: ModuleDeclarations,
     compiled_functions: SecondaryMap<FuncId, Option<CompiledBlob>>,
     compiled_data_objects: SecondaryMap<DataId, Option<CompiledBlob>>,
+    code_ranges: Vec<(usize, usize, FuncId)>,
     functions_to_finalize: Vec<FuncId>,
     data_objects_to_finalize: Vec<DataId>,
 }
 
-/// A handle to allow freeing memory allocated by the `Module`.
-struct MemoryHandle {
-    code: Memory,
-    readonly: Memory,
-    writable: Memory,
-}
-
 impl JITModule {
     /// Free memory allocated for code and data segments of compiled functions.
     ///
@@ -184,9 +191,7 @@ impl JITModule {
     /// from that module are currently executing and none of the `fn` pointers
     /// are called afterwards.
     pub unsafe fn free_memory(mut self) {
-        self.memory.code.free_memory();
-        self.memory.readonly.free_memory();
-        self.memory.writable.free_memory();
+        self.memory.free_memory();
     }
 
     fn lookup_symbol(&self, name: &str) -> Option<*const u8> {
@@ -207,7 +212,7 @@ impl JITModule {
     }
 
     fn get_address(&self, name: &ModuleRelocTarget) -> *const u8 {
-        match *name {
+        match name {
             ModuleRelocTarget::User { .. } => {
                 let (name, linkage) = if ModuleDeclarations::is_function(name) {
                     let func_id = FuncId::from_name(name);
@@ -239,12 +244,18 @@ impl JITModule {
                     panic!("can't resolve symbol {name}");
                 }
             }
-            ModuleRelocTarget::LibCall(ref libcall) => {
+            ModuleRelocTarget::LibCall(libcall) => {
                 let sym = (self.libcall_names)(*libcall);
                 self.lookup_symbol(&sym)
                     .unwrap_or_else(|| panic!("can't resolve libcall {sym}"))
             }
-            _ => panic!("invalid name"),
+            ModuleRelocTarget::FunctionOffset(func_id, offset) => {
+                match &self.compiled_functions[*func_id] {
+                    Some(compiled) => return compiled.ptr.wrapping_add(*offset as usize),
+                    None => todo!(),
+                }
+            }
+            name => panic!("invalid name {name:?}"),
         }
     }
 
@@ -324,9 +335,16 @@ impl JITModule {
             data.perform_relocations(|name| self.get_address(name));
         }
 
+        self.code_ranges
+            .sort_unstable_by_key(|(start, _end, _)| *start);
+
         // Now that we're done patching, prepare the memory for execution!
-        self.memory.readonly.set_readonly()?;
-        self.memory.code.set_readable_and_executable()?;
+        let branch_protection = if cfg!(target_arch = "aarch64") && use_bti(&self.isa.isa_flags()) {
+            BranchProtection::BTI
+        } else {
+            BranchProtection::None
+        };
+        self.memory.finalize(branch_protection)?;
 
         Ok(())
     }
@@ -338,30 +356,62 @@ impl JITModule {
             "cranelift-jit needs is_pic=false"
         );
 
-        let branch_protection =
-            if cfg!(target_arch = "aarch64") && use_bti(&builder.isa.isa_flags()) {
-                BranchProtection::BTI
-            } else {
-                BranchProtection::None
-            };
+        let memory = builder
+            .memory
+            .unwrap_or_else(|| Box::new(SystemMemoryProvider::new()));
         Self {
             isa: builder.isa,
             symbols: RefCell::new(builder.symbols),
             lookup_symbols: builder.lookup_symbols,
             libcall_names: builder.libcall_names,
-            memory: MemoryHandle {
-                code: Memory::new(branch_protection),
-                // Branch protection is not applicable to non-executable memory.
-                readonly: Memory::new(BranchProtection::None),
-                writable: Memory::new(BranchProtection::None),
-            },
+            memory,
             declarations: ModuleDeclarations::default(),
             compiled_functions: SecondaryMap::new(),
             compiled_data_objects: SecondaryMap::new(),
+            code_ranges: Vec::new(),
             functions_to_finalize: Vec::new(),
             data_objects_to_finalize: Vec::new(),
         }
     }
+
+    /// Look up the Wasmtime unwind ExceptionTable and corresponding
+    /// base PC, if any, for a given PC that may be within one of the
+    /// CompiledBlobs in this module.
+    #[cfg(feature = "wasmtime-unwinder")]
+    pub fn lookup_wasmtime_exception_data<'a>(
+        &'a self,
+        pc: usize,
+    ) -> Option<(usize, wasmtime_unwinder::ExceptionTable<'a>)> {
+        // Search the sorted code-ranges for the PC.
+        let idx = match self
+            .code_ranges
+            .binary_search_by_key(&pc, |(start, _end, _func)| *start)
+        {
+            Ok(exact_start_match) => Some(exact_start_match),
+            Err(least_upper_bound) if least_upper_bound > 0 => {
+                let last_range_before_pc = &self.code_ranges[least_upper_bound - 1];
+                if last_range_before_pc.0 <= pc && pc < last_range_before_pc.1 {
+                    Some(least_upper_bound - 1)
+                } else {
+                    None
+                }
+            }
+            _ => None,
+        }?;
+
+        let (start, _, func) = self.code_ranges[idx];
+
+        // Get the ExceptionTable. The "parse" here simply reads two
+        // u32s for lengths and constructs borrowed slices, so it's
+        // cheap.
+        let data = self.compiled_functions[func]
+            .as_ref()
+            .unwrap()
+            .exception_data
+            .as_ref()?;
+        let exception_table = wasmtime_unwinder::ExceptionTable::parse(data).ok()?;
+        Some((start, exception_table))
+    }
 }
 
 impl Module for JITModule {
@@ -436,15 +486,16 @@ impl Module for JITModule {
         let compiled_code = ctx.compiled_code().unwrap();
 
         let size = compiled_code.code_info().total_size as usize;
-        let align = alignment.max(self.isa.symbol_alignment());
-        let ptr = self
-            .memory
-            .code
-            .allocate(size, align)
-            .map_err(|e| ModuleError::Allocation {
-                message: "unable to alloc function",
-                err: e,
-            })?;
+        let align = alignment
+            .max(self.isa.function_alignment().minimum as u64)
+            .max(self.isa.symbol_alignment());
+        let ptr =
+            self.memory
+                .allocate_readexec(size, align)
+                .map_err(|e| ModuleError::Allocation {
+                    message: "unable to alloc function",
+                    err: e,
+                })?;
 
         {
             let mem = unsafe { std::slice::from_raw_parts_mut(ptr, size) };
@@ -459,7 +510,32 @@ impl Module for JITModule {
             .collect();
 
         self.record_function_for_perf(ptr, size, &decl.linkage_name(id));
-        self.compiled_functions[id] = Some(CompiledBlob { ptr, size, relocs });
+        self.compiled_functions[id] = Some(CompiledBlob {
+            ptr,
+            size,
+            relocs,
+            #[cfg(feature = "wasmtime-unwinder")]
+            exception_data: None,
+        });
+
+        let range_start = ptr as usize;
+        let range_end = range_start + size;
+        // These will be sorted when we finalize.
+        self.code_ranges.push((range_start, range_end, id));
+
+        #[cfg(feature = "wasmtime-unwinder")]
+        {
+            let mut exception_builder = wasmtime_unwinder::ExceptionTableBuilder::default();
+            exception_builder
+                .add_func(0, compiled_code.buffer.call_sites())
+                .map_err(|_| {
+                    ModuleError::Compilation(cranelift_codegen::CodegenError::Unsupported(
+                        "Invalid exception data".into(),
+                    ))
+                })?;
+            self.compiled_functions[id].as_mut().unwrap().exception_data =
+                Some(exception_builder.to_vec());
+        }
 
         self.functions_to_finalize.push(id);
 
@@ -473,7 +549,7 @@ impl Module for JITModule {
         bytes: &[u8],
         relocs: &[ModuleReloc],
     ) -> ModuleResult<()> {
-        info!("defining function {} with bytes", id);
+        info!("defining function {id} with bytes");
         let decl = self.declarations.get_function_decl(id);
         if !decl.linkage.is_definable() {
             return Err(ModuleError::InvalidImportDefinition(
@@ -488,15 +564,16 @@ impl Module for JITModule {
         }
 
         let size = bytes.len();
-        let align = alignment.max(self.isa.symbol_alignment());
-        let ptr = self
-            .memory
-            .code
-            .allocate(size, align)
-            .map_err(|e| ModuleError::Allocation {
-                message: "unable to alloc function bytes",
-                err: e,
-            })?;
+        let align = alignment
+            .max(self.isa.function_alignment().minimum as u64)
+            .max(self.isa.symbol_alignment());
+        let ptr =
+            self.memory
+                .allocate_readexec(size, align)
+                .map_err(|e| ModuleError::Allocation {
+                    message: "unable to alloc function bytes",
+                    err: e,
+                })?;
 
         unsafe {
             ptr::copy_nonoverlapping(bytes.as_ptr(), ptr, size);
@@ -507,6 +584,8 @@ impl Module for JITModule {
             ptr,
             size,
             relocs: relocs.to_owned(),
+            #[cfg(feature = "wasmtime-unwinder")]
+            exception_data: None,
         });
 
         self.functions_to_finalize.push(id);
@@ -538,6 +617,7 @@ impl Module for JITModule {
             data_relocs: _,
             custom_segment_section: _,
             align,
+            used: _,
         } = data;
 
         // Make sure to allocate at least 1 byte. Allocating 0 bytes is UB. Previously a dummy
@@ -548,16 +628,14 @@ impl Module for JITModule {
 
         let ptr = if decl.writable {
             self.memory
-                .writable
-                .allocate(alloc_size, align.unwrap_or(WRITABLE_DATA_ALIGNMENT))
+                .allocate_readwrite(alloc_size, align.unwrap_or(WRITABLE_DATA_ALIGNMENT))
                 .map_err(|e| ModuleError::Allocation {
                     message: "unable to alloc writable data",
                     err: e,
                 })?
         } else {
             self.memory
-                .readonly
-                .allocate(alloc_size, align.unwrap_or(READONLY_DATA_ALIGNMENT))
+                .allocate_readonly(alloc_size, align.unwrap_or(READONLY_DATA_ALIGNMENT))
                 .map_err(|e| ModuleError::Allocation {
                     message: "unable to alloc readonly data",
                     err: e,
@@ -599,6 +677,8 @@ impl Module for JITModule {
             ptr,
             size: init.size(),
             relocs,
+            #[cfg(feature = "wasmtime-unwinder")]
+            exception_data: None,
         });
         self.data_objects_to_finalize.push(id);
 
@@ -682,3 +762,14 @@ fn use_bti(isa_flags: &Vec<settings::Value>) -> bool {
         .find(|&f| f.name == "use_bti")
         .map_or(false, |f| f.as_bool().unwrap_or(false))
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_jit_module_is_send() {
+        fn assert_is_send<T: Send>() {}
+        assert_is_send::<JITModule>();
+    }
+}
