@@ -4,7 +4,7 @@ macro_rules! foreach_builtin_function {
     ($mac:ident) => {
         $mac! {
             // Returns an index for wasm's `memory.grow` builtin function.
-            memory32_grow(vmctx: vmctx, delta: u64, index: u32) -> pointer;
+            memory_grow(vmctx: vmctx, delta: u64, index: u32) -> pointer;
             // Returns an index for wasm's `table.copy` when both tables are locally
             // defined.
             table_copy(vmctx: vmctx, dst_index: u32, src_index: u32, dst: u64, src: u64, len: u64) -> bool;
@@ -74,11 +74,11 @@ macro_rules! foreach_builtin_function {
             #[cfg(feature = "gc-drc")]
             drop_gc_ref(vmctx: vmctx, val: u32);
 
-            // Do a GC, treating the optional `root` as a GC root and returning
-            // the updated `root` (so that, in the case of moving collectors,
-            // callers have a valid version of `root` again).
-            #[cfg(feature = "gc-drc")]
-            gc(vmctx: vmctx, root: u32) -> u64;
+            // Grow the GC heap by `bytes_needed` bytes.
+            //
+            // Traps if growing the GC heap fails.
+            #[cfg(feature = "gc-null")]
+            grow_gc_heap(vmctx: vmctx, bytes_needed: u64) -> bool;
 
             // Allocate a new, uninitialized GC object and return a reference to
             // it.
@@ -197,6 +197,20 @@ macro_rules! foreach_builtin_function {
             #[cfg(feature = "gc")]
             table_fill_gc_ref(vmctx: vmctx, table: u32, dst: u64, val: u32, len: u64) -> bool;
 
+            // Wasm floating-point routines for when the CPU instructions aren't available.
+            ceil_f32(vmctx: vmctx, x: f32) -> f32;
+            ceil_f64(vmctx: vmctx, x: f64) -> f64;
+            floor_f32(vmctx: vmctx, x: f32) -> f32;
+            floor_f64(vmctx: vmctx, x: f64) -> f64;
+            trunc_f32(vmctx: vmctx, x: f32) -> f32;
+            trunc_f64(vmctx: vmctx, x: f64) -> f64;
+            nearest_f32(vmctx: vmctx, x: f32) -> f32;
+            nearest_f64(vmctx: vmctx, x: f64) -> f64;
+            i8x16_swizzle(vmctx: vmctx, a: i8x16, b: i8x16) -> i8x16;
+            i8x16_shuffle(vmctx: vmctx, a: i8x16, b: i8x16, c: i8x16) -> i8x16;
+            fma_f32x4(vmctx: vmctx, x: f32x4, y: f32x4, z: f32x4) -> f32x4;
+            fma_f64x2(vmctx: vmctx, x: f64x2, y: f64x2, z: f64x2) -> f64x2;
+
             // Raises an unconditional trap with the specified code.
             //
             // This is used when signals-based-traps are disabled for backends
@@ -206,6 +220,33 @@ macro_rules! foreach_builtin_function {
             // Raises an unconditional trap where the trap information must have
             // been previously filled in.
             raise(vmctx: vmctx);
+
+            // Creates a new continuation from a funcref.
+            #[cfg(feature = "stack-switching")]
+            cont_new(vmctx: vmctx, r: pointer, param_count: u32, result_count: u32) -> pointer;
+
+            // Returns an index for Wasm's `table.grow` instruction
+            // for `contobj`s.  Note that the initial
+            // Option<VMContObj> (i.e., the value to fill the new
+            // slots with) is split into two arguments: The underlying
+            // continuation reference and the revision count.  To
+            // denote the continuation being `None`, `init_contref`
+            // may be 0.
+            #[cfg(feature = "stack-switching")]
+            table_grow_cont_obj(vmctx: vmctx, table: u32, delta: u64, init_contref: pointer, init_revision: size) -> pointer;
+
+            // `value_contref` and `value_revision` together encode
+            // the Option<VMContObj>, as in previous libcall.
+            #[cfg(feature = "stack-switching")]
+            table_fill_cont_obj(vmctx: vmctx, table: u32, dst: u64, value_contref: pointer, value_revision: size, len: u64) -> bool;
+
+            // Return the instance ID for a given vmctx.
+            #[cfg(feature = "gc")]
+            get_instance_id(vmctx: vmctx) -> u32;
+
+            // Throw an exception.
+            #[cfg(feature = "gc")]
+            throw_ref(vmctx: vmctx, exnref: u32) -> bool;
         }
     };
 }
@@ -214,11 +255,23 @@ macro_rules! foreach_builtin_function {
 /// `ComponentBuiltinFunctionIndex` using the iterator macro, e.g.
 /// `foreach_builtin_function`, as the way to generate accessor methods.
 macro_rules! declare_builtin_index {
-    ($index_name:ident, $iter:ident) => {
-        /// An index type for builtin functions.
-        #[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
+    (
+        $(#[$attr:meta])*
+        pub struct $index_name:ident : $for_each_builtin:ident ;
+    ) => {
+        $(#[$attr])*
+        #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
         pub struct $index_name(u32);
 
+        impl core::fmt::Debug for $index_name {
+            fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
+                f.debug_struct(stringify!($index_name))
+                    .field("index", &self.0)
+                    .field("ctor", &self.ctor_name())
+                    .finish()
+            }
+        }
+
         impl $index_name {
             /// Create a new builtin from its raw index
             pub const fn from_u32(i: u32) -> Self {
@@ -231,11 +284,42 @@ macro_rules! declare_builtin_index {
                 self.0
             }
 
-            $iter!(declare_builtin_index_constructors);
+            $for_each_builtin!(define_ctor_name);
+
+            $for_each_builtin!(declare_builtin_index_constructors);
+        }
+
+        #[cfg(test)]
+        impl arbitrary::Arbitrary<'_> for $index_name {
+            fn arbitrary(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<Self> {
+                Ok(Self(u.int_in_range(0..=Self::len() - 1)?))
+            }
         }
     };
 }
 
+/// Helper macro used by the above macro.
+macro_rules! define_ctor_name {
+    (
+        $(
+            $( #[$attr:meta] )*
+                $name:ident( $( $pname:ident: $param:ident ),* ) $( -> $result:ident )?;
+        )*
+    ) => {
+        /// Returns the name of the constructor that creates this index.
+        pub fn ctor_name(&self) -> &'static str {
+            let mut _i = self.0;
+            $(
+                if _i == 0 {
+                    return stringify!($name);
+                }
+                _i -= 1;
+            )*
+            unreachable!()
+        }
+    }
+}
+
 /// Helper macro used by the above macro.
 macro_rules! declare_builtin_index_constructors {
     (
@@ -299,7 +383,10 @@ macro_rules! declare_builtin_index_constructors {
 }
 
 // Define `struct BuiltinFunctionIndex`
-declare_builtin_index!(BuiltinFunctionIndex, foreach_builtin_function);
+declare_builtin_index! {
+    /// An index type for builtin functions.
+    pub struct BuiltinFunctionIndex : foreach_builtin_function;
+}
 
 /// Return value of [`BuiltinFunctionIndex::trap_sentinel`].
 pub enum TrapSentinel {
@@ -344,9 +431,10 @@ impl BuiltinFunctionIndex {
             }};
 
             // Growth-related functions return -2 as a sentinel.
-            (@get memory32_grow pointer) => (TrapSentinel::NegativeTwo);
+            (@get memory_grow pointer) => (TrapSentinel::NegativeTwo);
             (@get table_grow_func_ref pointer) => (TrapSentinel::NegativeTwo);
             (@get table_grow_gc_ref pointer) => (TrapSentinel::NegativeTwo);
+            (@get table_grow_cont_obj pointer) => (TrapSentinel::NegativeTwo);
 
             // Atomics-related functions return a negative value indicating trap
             // indicate a trap.
@@ -373,6 +461,22 @@ impl BuiltinFunctionIndex {
             (@get get_interned_func_ref pointer) => (return None);
             (@get intern_func_ref_for_gc_heap u64) => (return None);
             (@get is_subtype u32) => (return None);
+            (@get ceil_f32 f32) => (return None);
+            (@get ceil_f64 f64) => (return None);
+            (@get floor_f32 f32) => (return None);
+            (@get floor_f64 f64) => (return None);
+            (@get trunc_f32 f32) => (return None);
+            (@get trunc_f64 f64) => (return None);
+            (@get nearest_f32 f32) => (return None);
+            (@get nearest_f64 f64) => (return None);
+            (@get i8x16_swizzle i8x16) => (return None);
+            (@get i8x16_shuffle i8x16) => (return None);
+            (@get fma_f32x4 f32x4) => (return None);
+            (@get fma_f64x2 f64x2) => (return None);
+
+            (@get cont_new pointer) => (TrapSentinel::Negative);
+
+            (@get get_instance_id u32) => (return None);
 
             // Bool-returning functions use `false` as an indicator of a trap.
             (@get $name:ident bool) => (TrapSentinel::Falsy);
