@@ -6,6 +6,7 @@
 //      limits: *const VMStoreContext,
 //      flags: [VMGlobalDefinition; component.num_runtime_component_instances],
 //      trampoline_func_refs: [VMFuncRef; component.num_trampolines],
+//      unsafe_intrinsics: [VMFuncRef; component.num_unsafe_intrinsics],
 //      lowerings: [VMLowering; component.num_lowerings],
 //      memories: [*mut VMMemoryDefinition; component.num_runtime_memories],
 //      tables: [VMTable; component.num_runtime_tables],
@@ -14,8 +15,8 @@
 //      resource_destructors: [*mut VMFuncRef; component.num_resources],
 // }
 
-use crate::component::*;
 use crate::PtrSize;
+use crate::component::*;
 
 /// Equivalent of `VMCONTEXT_MAGIC` except for components.
 ///
@@ -58,6 +59,9 @@ pub struct VMComponentOffsets<P> {
     pub num_runtime_component_instances: u32,
     /// Number of cranelift-compiled trampolines required for this component.
     pub num_trampolines: u32,
+    /// Number of `VMFuncRef`s for unsafe intrinsics within this component's
+    /// context.
+    pub num_unsafe_intrinsics: u32,
     /// Number of resources within a component which need destructors stored.
     pub num_resources: u32,
 
@@ -67,6 +71,7 @@ pub struct VMComponentOffsets<P> {
     vm_store_context: u32,
     flags: u32,
     trampoline_func_refs: u32,
+    intrinsic_func_refs: u32,
     lowerings: u32,
     memories: u32,
     tables: u32,
@@ -90,22 +95,42 @@ impl<P: PtrSize> VMComponentOffsets<P> {
         let mut ret = Self {
             ptr,
             num_lowerings: component.num_lowerings,
-            num_runtime_memories: component.num_runtime_memories.try_into().unwrap(),
-            num_runtime_tables: component.num_runtime_tables.try_into().unwrap(),
-            num_runtime_reallocs: component.num_runtime_reallocs.try_into().unwrap(),
-            num_runtime_callbacks: component.num_runtime_callbacks.try_into().unwrap(),
-            num_runtime_post_returns: component.num_runtime_post_returns.try_into().unwrap(),
-            num_runtime_component_instances: component
-                .num_runtime_component_instances
-                .try_into()
-                .unwrap(),
+            num_runtime_memories: component.num_runtime_memories,
+            num_runtime_tables: component.num_runtime_tables,
+            num_runtime_reallocs: component.num_runtime_reallocs,
+            num_runtime_callbacks: component.num_runtime_callbacks,
+            num_runtime_post_returns: component.num_runtime_post_returns,
+            num_runtime_component_instances: component.num_runtime_component_instances,
             num_trampolines: component.trampolines.len().try_into().unwrap(),
+            num_unsafe_intrinsics: if let Some(i) = component
+                .unsafe_intrinsics
+                .iter()
+                .rposition(|x| x.is_some())
+            {
+                // Note: We do not currently have an indirection between "the
+                // `i`th unsafe intrinsic in the vmctx" and
+                // `UnsafeIntrinsic::from_u32(i)`, so therefore if we are
+                // compiling in *any* intrinsics, we need to include space for
+                // all of them up to the max `i` that is used.
+                //
+                // We _could_ introduce such an indirection via a map in
+                // `Component` like `PrimaryMap<UnsafeIntrinsicIndex,
+                // UnsafeIntrinsic>`, and that would allow us to densely pack
+                // intrinsics in the vmctx. However we do not do that today
+                // because there are very few unsafe intrinsics, and we do not
+                // see that changing anytime soon, so we aren't wasting much
+                // space.
+                u32::try_from(i + 1).unwrap()
+            } else {
+                0
+            },
             num_resources: component.num_resources,
             magic: 0,
             builtins: 0,
             vm_store_context: 0,
             flags: 0,
             trampoline_func_refs: 0,
+            intrinsic_func_refs: 0,
             lowerings: 0,
             memories: 0,
             tables: 0,
@@ -149,9 +174,10 @@ impl<P: PtrSize> VMComponentOffsets<P> {
             size(flags) = cmul(ret.num_runtime_component_instances, ret.ptr.size_of_vmglobal_definition()),
             align(u32::from(ret.ptr.size())),
             size(trampoline_func_refs) = cmul(ret.num_trampolines, ret.ptr.size_of_vm_func_ref()),
+            size(intrinsic_func_refs) = cmul(ret.num_unsafe_intrinsics, ret.ptr.size_of_vm_func_ref()),
             size(lowerings) = cmul(ret.num_lowerings, ret.ptr.size() * 2),
             size(memories) = cmul(ret.num_runtime_memories, ret.ptr.size()),
-            size(tables) = cmul(ret.num_runtime_tables, ret.size_of_vmtable()),
+            size(tables) = cmul(ret.num_runtime_tables, ret.size_of_vmtable_import()),
             size(reallocs) = cmul(ret.num_runtime_reallocs, ret.ptr.size()),
             size(callbacks) = cmul(ret.num_runtime_callbacks, ret.ptr.size()),
             size(post_returns) = cmul(ret.num_runtime_post_returns, ret.ptr.size()),
@@ -212,6 +238,20 @@ impl<P: PtrSize> VMComponentOffsets<P> {
         self.trampoline_func_refs() + index.as_u32() * u32::from(self.ptr.size_of_vm_func_ref())
     }
 
+    /// The offset of the `unsafe_intrinsic_func_refs` field.
+    #[inline]
+    pub fn unsafe_intrinsic_func_refs(&self) -> u32 {
+        self.intrinsic_func_refs
+    }
+
+    /// The offset of the `VMFuncRef` for the `intrinsic` specified.
+    #[inline]
+    pub fn unsafe_intrinsic_func_ref(&self, intrinsic: UnsafeIntrinsic) -> u32 {
+        assert!(intrinsic.index() < self.num_unsafe_intrinsics);
+        self.unsafe_intrinsic_func_refs()
+            + intrinsic.index() * u32::from(self.ptr.size_of_vm_func_ref())
+    }
+
     /// The offset of the `lowerings` field.
     #[inline]
     pub fn lowerings(&self) -> u32 {
@@ -279,14 +319,14 @@ impl<P: PtrSize> VMComponentOffsets<P> {
     #[inline]
     pub fn runtime_table(&self, index: RuntimeTableIndex) -> u32 {
         assert!(index.as_u32() < self.num_runtime_tables);
-        self.runtime_tables() + index.as_u32() * u32::from(self.size_of_vmtable())
+        self.runtime_tables() + index.as_u32() * u32::from(self.size_of_vmtable_import())
     }
 
-    /// Return the size of `VMTable`, used here to hold the pointers to
+    /// Return the size of `VMTableImport`, used here to hold the pointers to
     /// the `VMTableDefinition` and `VMContext`.
     #[inline]
-    pub fn size_of_vmtable(&self) -> u8 {
-        2 * self.pointer_size()
+    pub fn size_of_vmtable_import(&self) -> u8 {
+        3 * self.pointer_size()
     }
 
     /// The offset of the base of the `runtime_reallocs` field
