@@ -308,6 +308,7 @@ impl Config {
             Some(self.module_config.config.shared_everything_threads_enabled);
         cfg.wasm.wide_arithmetic = Some(self.module_config.config.wide_arithmetic_enabled);
         cfg.wasm.exceptions = Some(self.module_config.config.exceptions_enabled);
+        cfg.wasm.shared_memory = Some(self.module_config.shared_memory);
         if !self.module_config.config.simd_enabled {
             cfg.wasm.relaxed_simd = Some(false);
         }
@@ -448,6 +449,15 @@ impl Config {
             self.wasmtime.async_config.configure(&mut cfg);
         }
 
+        // Fuzzing on macOS with mach ports seems to sometimes bypass the mach
+        // port handling thread entirely and go straight to asan's or fuzzing's
+        // signal handler. No idea why and for me at least it's just easier to
+        // disable mach ports when fuzzing because there's no need to use that
+        // over signal handlers.
+        if cfg!(target_vendor = "apple") {
+            cfg.macos_use_mach_ports(false);
+        }
+
         return cfg;
     }
 
