@@ -10,6 +10,7 @@
 //! When an oracle finds a bug, it should report it to the fuzzing engine by
 //! panicking.
 
+pub mod component_api;
 #[cfg(feature = "fuzz-spec-interpreter")]
 pub mod diff_spec;
 pub mod diff_wasmi;
@@ -21,6 +22,8 @@ mod stacks;
 
 use self::diff_wasmtime::WasmtimeInstance;
 use self::engine::{DiffEngine, DiffInstance};
+use crate::block_on;
+use crate::generators::GcOps;
 use crate::generators::{self, CompilerStrategy, DiffValue, DiffValueType};
 use crate::single_module_fuzzer::KnownValid;
 use arbitrary::Arbitrary;
@@ -29,7 +32,7 @@ use std::future::Future;
 use std::pin::Pin;
 use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering::SeqCst};
 use std::sync::{Arc, Condvar, Mutex};
-use std::task::{Context, Poll, Waker};
+use std::task::{Context, Poll};
 use std::time::{Duration, Instant};
 use wasmtime::*;
 use wasmtime_wast::WastContext;
@@ -58,7 +61,10 @@ pub fn log_wasm(wasm: &[u8]) {
         // If wasmprinter failed remove a `*.wat` file, if any, to avoid
         // confusing a preexisting one with this wasm which failed to get
         // printed.
-        Err(_) => drop(std::fs::remove_file(&wat)),
+        Err(e) => {
+            log::debug!("failed to print to wat: {e}");
+            drop(std::fs::remove_file(&wat))
+        }
     }
 }
 
@@ -703,6 +709,7 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
     crate::init_fuzzing();
 
     let mut fuzz_config: generators::Config = u.arbitrary()?;
+    fuzz_config.module_config.shared_memory = true;
     let test: generators::WastTest = u.arbitrary()?;
 
     let test = &test.test;
@@ -775,14 +782,11 @@ pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
     Ok(())
 }
 
-/// Execute a series of `table.get` and `table.set` operations.
+/// Execute a series of `gc` operations.
 ///
 /// Returns the number of `gc` operations which occurred throughout the test
 /// case -- used to test below that gc happens reasonably soon and eventually.
-pub fn table_ops(
-    mut fuzz_config: generators::Config,
-    mut ops: generators::table_ops::TableOps,
-) -> Result<usize> {
+pub fn gc_ops(mut fuzz_config: generators::Config, mut ops: GcOps) -> Result<usize> {
     let expected_drops = Arc::new(AtomicUsize::new(0));
     let num_dropped = Arc::new(AtomicUsize::new(0));
 
@@ -815,7 +819,7 @@ pub fn table_ops(
             let expected_drops = expected_drops.clone();
             let num_gcs = num_gcs.clone();
             move |mut caller: Caller<'_, StoreLimits>, _params, results| {
-                log::info!("table_ops: GC");
+                log::info!("gc_ops: GC");
                 if num_gcs.fetch_add(1, SeqCst) < MAX_GCS {
                     caller.gc(None);
                 }
@@ -833,7 +837,7 @@ pub fn table_ops(
                     CountDrops::new(&expected_drops, num_dropped.clone()),
                 )?;
 
-                log::info!("table_ops: gc() -> ({a:?}, {b:?}, {c:?})");
+                log::info!("gc_ops: gc() -> ({a:?}, {b:?}, {c:?})");
                 results[0] = Some(a).into();
                 results[1] = Some(b).into();
                 results[2] = Some(c).into();
@@ -850,7 +854,7 @@ pub fn table_ops(
                       b: Option<Rooted<ExternRef>>,
                       c: Option<Rooted<ExternRef>>|
                       -> Result<()> {
-                    log::info!("table_ops: take_refs({a:?}, {b:?}, {c:?})",);
+                    log::info!("gc_ops: take_refs({a:?}, {b:?}, {c:?})",);
 
                     // Do the assertion on each ref's inner data, even though it
                     // all points to the same atomic, so that if we happen to
@@ -894,7 +898,7 @@ pub fn table_ops(
             let num_dropped = num_dropped.clone();
             let expected_drops = expected_drops.clone();
             move |mut caller, _params, results| {
-                log::info!("table_ops: make_refs");
+                log::info!("gc_ops: make_refs");
 
                 let a = ExternRef::new(
                     &mut caller,
@@ -909,7 +913,7 @@ pub fn table_ops(
                     CountDrops::new(&expected_drops, num_dropped.clone()),
                 )?;
 
-                log::info!("table_ops: make_refs() -> ({a:?}, {b:?}, {c:?})");
+                log::info!("gc_ops: make_refs() -> ({a:?}, {b:?}, {c:?})");
 
                 results[0] = Some(a).into();
                 results[1] = Some(b).into();
@@ -920,6 +924,38 @@ pub fn table_ops(
         });
         linker.define(&store, "", "make_refs", func).unwrap();
 
+        let func_ty = FuncType::new(
+            store.engine(),
+            vec![ValType::Ref(RefType::new(false, HeapType::Any))],
+            vec![],
+        );
+
+        let func = Func::new(&mut store, func_ty, {
+            move |_caller: Caller<'_, StoreLimits>, _params, _results| {
+                log::info!("gc_ops: take_struct(<ref any>)");
+                Ok(())
+            }
+        });
+
+        linker.define(&store, "", "take_struct", func).unwrap();
+
+        for imp in module.imports() {
+            if imp.module() == "" {
+                let name = imp.name();
+                if name.starts_with("take_struct_") {
+                    if let wasmtime::ExternType::Func(ft) = imp.ty() {
+                        let imp_name = name.to_string();
+                        let func =
+                            Func::new(&mut store, ft.clone(), move |_caller, _params, _results| {
+                                log::info!("gc_ops: {imp_name}(<typed structref>)");
+                                Ok(())
+                            });
+                        linker.define(&store, "", name, func).unwrap();
+                    }
+                }
+            }
+        }
+
         let instance = linker.instantiate(&mut store, &module).unwrap();
         let run = instance.get_func(&mut store, "run").unwrap();
 
@@ -927,7 +963,7 @@ pub fn table_ops(
             let mut scope = RootScope::new(&mut store);
 
             log::info!(
-                "table_ops: begin allocating {} externref arguments",
+                "gc_ops: begin allocating {} externref arguments",
                 ops.limits.num_globals
             );
             let args: Vec<_> = (0..ops.limits.num_params)
@@ -939,7 +975,7 @@ pub fn table_ops(
                 })
                 .collect::<Result<_>>()?;
             log::info!(
-                "table_ops: end allocating {} externref arguments",
+                "gc_ops: end allocating {} externref arguments",
                 ops.limits.num_globals
             );
 
@@ -947,18 +983,17 @@ pub fn table_ops(
             // valid traps are table-out-of-bounds which happens through `table.get`
             // and `table.set` generated or an out-of-fuel trap. Otherwise any other
             // error is unexpected and should fail fuzzing.
-            log::info!("table_ops: calling into Wasm `run` function");
+            log::info!("gc_ops: calling into Wasm `run` function");
             let err = run.call(&mut scope, &args, &mut []).unwrap_err();
-            match err.downcast::<GcHeapOutOfMemory<CountDrops>>() {
-                Ok(_oom) => {}
-                Err(err) => {
-                    let trap = err
-                        .downcast::<Trap>()
-                        .expect("if not GC oom, error should be a Wasm trap");
-                    match trap {
-                        Trap::TableOutOfBounds | Trap::OutOfFuel => {}
-                        _ => panic!("unexpected trap: {trap}"),
-                    }
+            if err.is::<GcHeapOutOfMemory<CountDrops>>() || err.is::<GcHeapOutOfMemory<()>>() {
+                // Accept GC OOM as an allowed outcome for this fuzzer.
+            } else {
+                let trap = err
+                    .downcast::<Trap>()
+                    .expect("if not GC oom, error should be a Wasm trap");
+                match trap {
+                    Trap::TableOutOfBounds | Trap::OutOfFuel | Trap::AllocationTooLarge => {}
+                    _ => panic!("unexpected trap: {trap}"),
                 }
             }
         }
@@ -1043,98 +1078,6 @@ impl Drop for HelperThread {
     }
 }
 
-/// Generate and execute a `crate::generators::component_types::TestCase` using the specified `input` to create
-/// arbitrary types and values.
-pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbitrary::Result<()> {
-    use crate::generators::component_types;
-    use wasmtime::component::{Component, Linker, Val};
-    use wasmtime_test_util::component::FuncExt;
-    use wasmtime_test_util::component_fuzz::{
-        EXPORT_FUNCTION, IMPORT_FUNCTION, MAX_TYPE_DEPTH, TestCase, Type,
-    };
-
-    crate::init_fuzzing();
-
-    let mut types = Vec::new();
-    let mut type_fuel = 500;
-
-    for _ in 0..5 {
-        types.push(Type::generate(input, MAX_TYPE_DEPTH, &mut type_fuel)?);
-    }
-    let params = (0..input.int_in_range(0..=5)?)
-        .map(|_| input.choose(&types))
-        .collect::<arbitrary::Result<Vec<_>>>()?;
-    let result = if input.arbitrary()? {
-        Some(input.choose(&types)?)
-    } else {
-        None
-    };
-
-    let case = TestCase {
-        params,
-        result,
-        encoding1: input.arbitrary()?,
-        encoding2: input.arbitrary()?,
-    };
-
-    let mut config = wasmtime_test_util::component::config();
-    config.debug_adapter_modules(input.arbitrary()?);
-    let engine = Engine::new(&config).unwrap();
-    let mut store = Store::new(&engine, (Vec::new(), None));
-    let wat = case.declarations().make_component();
-    let wat = wat.as_bytes();
-    log_wasm(wat);
-    let component = Component::new(&engine, wat).unwrap();
-    let mut linker = Linker::new(&engine);
-
-    linker
-        .root()
-        .func_new(IMPORT_FUNCTION, {
-            move |mut cx: StoreContextMut<'_, (Vec<Val>, Option<Vec<Val>>)>,
-                  _,
-                  params: &[Val],
-                  results: &mut [Val]|
-                  -> Result<()> {
-                log::trace!("received params {params:?}");
-                let (expected_args, expected_results) = cx.data_mut();
-                assert_eq!(params.len(), expected_args.len());
-                for (expected, actual) in expected_args.iter().zip(params) {
-                    assert_eq!(expected, actual);
-                }
-                results.clone_from_slice(&expected_results.take().unwrap());
-                log::trace!("returning results {results:?}");
-                Ok(())
-            }
-        })
-        .unwrap();
-
-    let instance = linker.instantiate(&mut store, &component).unwrap();
-    let func = instance.get_func(&mut store, EXPORT_FUNCTION).unwrap();
-    let ty = func.ty(&store);
-
-    while input.arbitrary()? {
-        let params = ty
-            .params()
-            .map(|(_, ty)| component_types::arbitrary_val(&ty, input))
-            .collect::<arbitrary::Result<Vec<_>>>()?;
-        let results = ty
-            .results()
-            .map(|ty| component_types::arbitrary_val(&ty, input))
-            .collect::<arbitrary::Result<Vec<_>>>()?;
-
-        *store.data_mut() = (params.clone(), Some(results.clone()));
-
-        log::trace!("passing params {params:?}");
-        let mut actual = vec![Val::Bool(false); results.len()];
-        func.call_and_post_return(&mut store, &params, &mut actual)
-            .unwrap();
-        log::trace!("received results {actual:?}");
-        assert_eq!(actual, results);
-    }
-
-    Ok(())
-}
-
 /// Instantiates a wasm module and runs its exports with dummy values, all in
 /// an async fashion.
 ///
@@ -1192,7 +1135,7 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
     // Run the instantiation process, asynchronously, and if everything
     // succeeds then pull out the instance.
     // log::info!("starting instantiation");
-    let instance = run(Timeout {
+    let instance = block_on(Timeout {
         future: Instance::new_async(&mut store, &module, &imports),
         polls: take_poll_amt(&mut poll_amts),
         end: Instant::now() + Duration::from_millis(2_000),
@@ -1238,7 +1181,7 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
 
         log::info!("invoking export {name:?}");
         let future = func.call_async(&mut store, &params, &mut results);
-        match run(Timeout {
+        match block_on(Timeout {
             future,
             polls: take_poll_amt(&mut poll_amts),
             end: Instant::now() + Duration::from_millis(2_000),
@@ -1324,67 +1267,20 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
             }
         }
     }
-
-    fn run<F: Future>(future: F) -> F::Output {
-        let mut f = Box::pin(future);
-        let mut cx = Context::from_waker(Waker::noop());
-        loop {
-            match f.as_mut().poll(&mut cx) {
-                Poll::Ready(val) => break val,
-                Poll::Pending => {}
-            }
-        }
-    }
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
-    use arbitrary::Unstructured;
-    use rand::prelude::*;
+    use crate::test::{gen_until_pass, test_n_times};
     use wasmparser::{Validator, WasmFeatures};
 
-    fn gen_until_pass<T: for<'a> Arbitrary<'a>>(
-        mut f: impl FnMut(T, &mut Unstructured<'_>) -> Result<bool>,
-    ) -> bool {
-        let mut rng = SmallRng::seed_from_u64(0);
-        let mut buf = vec![0; 2048];
-        let n = 3000;
-        for _ in 0..n {
-            rng.fill_bytes(&mut buf);
-            let mut u = Unstructured::new(&buf);
-
-            if let Ok(config) = u.arbitrary() {
-                if f(config, &mut u).unwrap() {
-                    return true;
-                }
-            }
-        }
-        false
-    }
-
-    /// Runs `f` with random data until it returns `Ok(())` `iters` times.
-    fn test_n_times<T: for<'a> Arbitrary<'a>>(
-        iters: u32,
-        mut f: impl FnMut(T, &mut Unstructured<'_>) -> arbitrary::Result<()>,
-    ) {
-        let mut to_test = 0..iters;
-        let ok = gen_until_pass(|a, b| {
-            if f(a, b).is_ok() {
-                Ok(to_test.next().is_none())
-            } else {
-                Ok(false)
-            }
-        });
-        assert!(ok);
-    }
-
-    // Test that the `table_ops` fuzzer eventually runs the gc function in the host.
+    // Test that the `gc_ops` fuzzer eventually runs the gc function in the host.
     // We've historically had issues where this fuzzer accidentally wasn't fuzzing
     // anything for a long time so this is an attempt to prevent that from happening
     // again.
     #[test]
-    fn table_ops_eventually_gcs() {
+    fn gc_ops_eventually_gcs() {
         // Skip if we're under emulation because some fuzz configurations will do
         // large address space reservations that QEMU doesn't handle well.
         if std::env::var("WASMTIME_TEST_NO_HOG_MEMORY").is_ok() {
@@ -1392,7 +1288,7 @@ mod tests {
         }
 
         let ok = gen_until_pass(|(config, test), _| {
-            let result = table_ops(config, test)?;
+            let result = gc_ops(config, test)?;
             Ok(result > 0)
         });
 
@@ -1415,7 +1311,6 @@ mod tests {
             | WasmFeatures::SIMD
             | WasmFeatures::MULTI_MEMORY
             | WasmFeatures::RELAXED_SIMD
-            | WasmFeatures::THREADS
             | WasmFeatures::TAIL_CALL
             | WasmFeatures::WIDE_ARITHMETIC
             | WasmFeatures::MEMORY64
