@@ -648,11 +648,7 @@ impl TransmitTest for DynamicTransmitTest {
                 .get_export_index(store.as_context_mut(), None, "local:local/transmit")
                 .ok_or_else(|| anyhow!("can't find `local:local/transmit` in instance"))?;
             let exchange_function = instance
-                .get_export_index(
-                    store.as_context_mut(),
-                    Some(&transmit_instance),
-                    "[async]exchange",
-                )
+                .get_export_index(store.as_context_mut(), Some(&transmit_instance), "exchange")
                 .ok_or_else(|| anyhow!("can't find `exchange` in instance"))?;
             instance
                 .get_func(store.as_context_mut(), exchange_function)
