@@ -52,6 +52,26 @@ pub fn find_tests(root: &Path) -> Result<Vec<WastTest>> {
         &FindConfig::Infer(component_test_config),
     )
     .with_context(|| format!("failed to add tests from `{}`", cm_tests.display()))?;
+
+    // Temporarily work around upstream tests that loop forever.
+    //
+    // Now that `thread.yield` and `CALLBACK_CODE_YIELD` are both no-ops in
+    // non-blocking contexts, these tests need to be updated; meanwhile, we skip
+    // them.
+    //
+    // TODO: remove this once
+    // https://github.com/WebAssembly/component-model/pull/578 has been merged:
+    {
+        let skip_list = &["drop-subtask.wast", "async-calls-sync.wast"];
+        tests.retain(|test| {
+            test.path
+                .file_name()
+                .and_then(|name| name.to_str())
+                .map(|name| !skip_list.contains(&name))
+                .unwrap_or(true)
+        });
+    }
+
     Ok(tests)
 }
 
@@ -168,6 +188,16 @@ fn component_test_config(test: &Path) -> TestConfig {
             ret.component_model_async = Some(true);
             ret.component_model_async_builtins = Some(true);
         }
+        if parent.ends_with("wasm-tools") {
+            ret.memory64 = Some(true);
+            ret.threads = Some(true);
+            ret.exceptions = Some(true);
+            ret.gc = Some(true);
+        }
+        if parent.ends_with("wasmtime") {
+            ret.exceptions = Some(true);
+            ret.gc = Some(true);
+        }
     }
 
     ret
@@ -667,16 +697,28 @@ impl WastTest {
         let failing_component_model_tests = [
             // FIXME(#11683)
             "component-model/test/values/trap-in-post-return.wast",
-            // Awaiting https://github.com/WebAssembly/component-model/pull/570
-            "component-model/test/resources/multiple-resources.wast",
-            "component-model/test/async/empty-wait.wast",
-            "component-model/test/async/drop-stream.wast",
-            "component-model/test/async/passing-resources.wast",
+            "component-model/test/wasmtime/resources.wast",
+            "component-model/test/wasm-tools/naming.wast",
+            // TODO: remove these once
+            // https://github.com/WebAssembly/component-model/pull/578 has been
+            // merged:
             "component-model/test/async/async-calls-sync.wast",
-            "component-model/test/async/partial-stream-copies.wast",
-            "component-model/test/async/futures-must-write.wast",
+            "component-model/test/async/backpressure-deadlock.wast",
             "component-model/test/async/cancel-stream.wast",
+            "component-model/test/async/cancel-subtask.wast",
+            "component-model/test/async/deadlock.wast",
+            "component-model/test/async/drop-subtask.wast",
             "component-model/test/async/drop-waitable-set.wast",
+            "component-model/test/async/empty-wait.wast",
+            "component-model/test/async/fused.wast",
+            "component-model/test/async/future-read.wast",
+            "component-model/test/async/partial-stream-copies.wast",
+            "component-model/test/async/passing-resources.wast",
+            "component-model/test/async/stackful.wast",
+            "component-model/test/async/trap-if-block-and-sync.wast",
+            "component-model/test/async/trap-if-done.wast",
+            "component-model/test/async/wait-during-callback.wast",
+            "component-model/test/async/zero-length.wast",
         ];
         if failing_component_model_tests
             .iter()
