@@ -70,6 +70,7 @@ pub(crate) fn build_module_artifacts<T: FinishedObject>(
     T,
     Option<(CompiledModuleInfo, CompiledFunctionsTable, ModuleTypes)>,
 )> {
+    let compiler = engine.try_compiler()?;
     let tunables = engine.tunables();
 
     // First a `ModuleEnvironment` is created which records type information
@@ -101,7 +102,7 @@ pub(crate) fn build_module_artifacts<T: FinishedObject>(
 
     // Emplace all compiled functions into the object file with any other
     // sections associated with code as well.
-    let mut object = engine.compiler().object(ObjectKind::Module)?;
+    let mut object = compiler.object(ObjectKind::Module)?;
     // Insert `Engine` and type-level information into the compiled
     // artifact so if this module is deserialized later it contains all
     // information necessary.
@@ -111,7 +112,7 @@ pub(crate) fn build_module_artifacts<T: FinishedObject>(
     // They're only used during deserialization and not during runtime for
     // the module itself. Currently there's no need for that, however, so
     // it's left as an exercise for later.
-    engine.append_compiler_info(&mut object);
+    engine.append_compiler_info(&mut object)?;
     engine.append_bti(&mut object);
 
     let (mut object, compilation_artifacts) = indices.link_and_append_code(
@@ -150,8 +151,8 @@ pub(crate) fn build_component_artifacts<T: FinishedObject>(
         CompiledComponentInfo, ComponentArtifacts, ComponentTypesBuilder,
     };
 
+    let compiler = engine.try_compiler()?;
     let tunables = engine.tunables();
-    let compiler = engine.compiler();
 
     let scope = ScopeVec::new();
     let mut validator = wasmparser::Validator::new_with_features(engine.features());
@@ -185,7 +186,7 @@ pub(crate) fn build_component_artifacts<T: FinishedObject>(
     }
 
     let mut object = compiler.object(ObjectKind::Component)?;
-    engine.append_compiler_info(&mut object);
+    engine.append_compiler_info(&mut object)?;
     engine.append_bti(&mut object);
 
     let (mut object, compilation_artifacts) = indices.link_and_append_code(
@@ -520,7 +521,7 @@ impl<'a> CompileInputs<'a> {
     /// Compile these `CompileInput`s (maybe in parallel) and return the
     /// resulting `UnlinkedCompileOutput`s.
     fn compile(self, engine: &Engine) -> Result<UnlinkedCompileOutputs<'a>> {
-        let compiler = engine.compiler();
+        let compiler = engine.try_compiler()?;
 
         if self.inputs.len() > 0 && cfg!(miri) {
             bail!(
@@ -912,7 +913,7 @@ the use case.
 }
 
 fn compile_required_builtins(engine: &Engine, raw_outputs: &mut Vec<CompileOutput>) -> Result<()> {
-    let compiler = engine.compiler();
+    let compiler = engine.try_compiler()?;
     let mut builtins = HashSet::new();
     let mut new_inputs: Vec<CompileInput<'_>> = Vec::new();
 
@@ -1040,7 +1041,7 @@ impl FunctionIndices {
         // The result is a vector parallel to `compiled_funcs` where
         // `symbol_ids_and_locs[i]` is the symbol ID and function location of
         // `compiled_funcs[i]`.
-        let compiler = engine.compiler();
+        let compiler = engine.try_compiler()?;
         let tunables = engine.tunables();
         let symbol_ids_and_locs = compiler.append_code(
             &mut obj,
