@@ -1,14 +1,15 @@
 //! Memory management for executable code.
 
-use crate::prelude::*;
-use crate::runtime::vm::{libcalls, MmapVec};
 use crate::Engine;
+use crate::prelude::*;
+use crate::runtime::vm::MmapVec;
 use alloc::sync::Arc;
 use core::ops::Range;
+use object::SectionFlags;
 use object::endian::Endianness;
-use object::read::{elf::ElfFile64, Object, ObjectSection};
-use object::{ObjectSymbol, SectionFlags};
-use wasmtime_environ::{lookup_trap_code, obj, Trap};
+use object::read::{Object, ObjectSection, elf::ElfFile64};
+use wasmtime_environ::{Trap, lookup_trap_code, obj};
+use wasmtime_unwinder::ExceptionTable;
 
 /// Management of executable memory within a `MmapVec`
 ///
@@ -27,14 +28,15 @@ pub struct CodeMemory {
     has_native_debug_info: bool,
     custom_code_memory: Option<Arc<dyn CustomCodeMemory>>,
 
-    relocations: Vec<(usize, obj::LibCall)>,
-
     // Ranges within `self.mmap` of where the particular sections lie.
     text: Range<usize>,
     unwind: Range<usize>,
     trap_data: Range<usize>,
     wasm_data: Range<usize>,
     address_map_data: Range<usize>,
+    stack_map_data: Range<usize>,
+    exception_data: Range<usize>,
+    frame_tables_data: Range<usize>,
     func_name_data: Range<usize>,
     info_data: Range<usize>,
     wasm_dwarf: Range<usize>,
@@ -45,9 +47,11 @@ impl Drop for CodeMemory {
         // If there is a custom code memory handler, restore the
         // original (non-executable) state of the memory.
         if let Some(mem) = self.custom_code_memory.as_ref() {
-            let text = self.text();
-            mem.unpublish_executable(text.as_ptr(), text.len())
-                .expect("Executable memory unpublish failed");
+            if self.published && self.needs_executable {
+                let text = self.text();
+                mem.unpublish_executable(text.as_ptr(), text.len())
+                    .expect("Executable memory unpublish failed");
+            }
         }
 
         // Drop the registrations before `self.mmap` since they (implicitly) refer to it.
@@ -111,7 +115,6 @@ impl CodeMemory {
             .map_err(obj::ObjectCrateErrorWrapper)
             .with_context(|| "failed to parse internal compilation artifact")?;
 
-        let mut relocations = Vec::new();
         let mut text = 0..0;
         let mut unwind = 0..0;
         let mut enable_branch_protection = None;
@@ -119,8 +122,11 @@ impl CodeMemory {
         #[cfg(feature = "debug-builtins")]
         let mut has_native_debug_info = false;
         let mut trap_data = 0..0;
+        let mut exception_data = 0..0;
+        let mut frame_tables_data = 0..0;
         let mut wasm_data = 0..0;
         let mut address_map_data = 0..0;
+        let mut stack_map_data = 0..0;
         let mut func_name_data = 0..0;
         let mut info_data = 0..0;
         let mut wasm_dwarf = 0..0;
@@ -154,34 +160,21 @@ impl CodeMemory {
                         }
                     }
 
-                    // The text section might have relocations for things like
-                    // libcalls which need to be applied, so handle those here.
-                    //
-                    // Note that only a small subset of possible relocations are
-                    // handled. Only those required by the compiler side of
-                    // things are processed.
-                    for (offset, reloc) in section.relocations() {
-                        assert_eq!(reloc.kind(), object::RelocationKind::Absolute);
-                        assert_eq!(reloc.encoding(), object::RelocationEncoding::Generic);
-                        assert_eq!(usize::from(reloc.size()), core::mem::size_of::<usize>() * 8);
-                        assert_eq!(reloc.addend(), 0);
-                        let sym = match reloc.target() {
-                            object::RelocationTarget::Symbol(id) => id,
-                            other => panic!("unknown relocation target {other:?}"),
-                        };
-                        let sym = obj.symbol_by_index(sym).unwrap().name().unwrap();
-                        let libcall = obj::LibCall::from_str(sym)
-                            .unwrap_or_else(|| panic!("unknown symbol relocation: {sym}"));
-
-                        let offset = usize::try_from(offset).unwrap();
-                        relocations.push((offset, libcall));
-                    }
+                    // Assert that Cranelift hasn't inserted any calls that need to be
+                    // relocated. We avoid using things like Cranelift's floor/ceil/etc.
+                    // operators in the Wasm-to-Cranelift translator specifically to
+                    // avoid having to do any relocations here. This also ensures that
+                    // all builtins use the same trampoline mechanism.
+                    assert!(section.relocations().next().is_none());
                 }
                 #[cfg(has_host_compiler_backend)]
                 crate::runtime::vm::UnwindRegistration::SECTION_NAME => unwind = range,
                 obj::ELF_WASM_DATA => wasm_data = range,
                 obj::ELF_WASMTIME_ADDRMAP => address_map_data = range,
+                obj::ELF_WASMTIME_STACK_MAP => stack_map_data = range,
                 obj::ELF_WASMTIME_TRAPS => trap_data = range,
+                obj::ELF_WASMTIME_EXCEPTIONS => exception_data = range,
+                obj::ELF_WASMTIME_FRAMES => frame_tables_data = range,
                 obj::ELF_NAME_DATA => func_name_data = range,
                 obj::ELF_WASMTIME_INFO => info_data = range,
                 obj::ELF_WASMTIME_DWARF => wasm_dwarf = range,
@@ -196,6 +189,17 @@ impl CodeMemory {
         #[cfg(not(has_host_compiler_backend))]
         let _ = &mut unwind;
 
+        // Ensure that the exception table is well-formed. This parser
+        // construction is cheap: it reads the header and validates
+        // ranges but nothing else. We do this only in debug-assertion
+        // builds because we otherwise require for safety that the
+        // compiled artifact is as-produced-by this version of
+        // Wasmtime, and we should always produce a correct exception
+        // table (i.e., we are not expecting untrusted data here).
+        if cfg!(debug_assertions) {
+            let _ = ExceptionTable::parse(&mmap[exception_data.clone()])?;
+        }
+
         Ok(Self {
             mmap,
             #[cfg(has_host_compiler_backend)]
@@ -213,11 +217,13 @@ impl CodeMemory {
             unwind,
             trap_data,
             address_map_data,
+            stack_map_data,
+            exception_data,
+            frame_tables_data,
             func_name_data,
             wasm_dwarf,
             info_data,
             wasm_data,
-            relocations,
         })
     }
 
@@ -263,6 +269,24 @@ impl CodeMemory {
         &self.mmap[self.address_map_data.clone()]
     }
 
+    /// Returns the encoded stack map section used to pass to
+    /// `wasmtime_environ::StackMap::lookup`.
+    pub fn stack_map_data(&self) -> &[u8] {
+        &self.mmap[self.stack_map_data.clone()]
+    }
+
+    /// Returns the encoded exception-tables section to pass to
+    /// `wasmtime_unwinder::ExceptionTable::parse`.
+    pub fn exception_tables(&self) -> &[u8] {
+        &self.mmap[self.exception_data.clone()]
+    }
+
+    /// Returns the encoded frame-tables section to pass to
+    /// `wasmtime_environ::FrameTable::parse`.
+    pub fn frame_tables(&self) -> &[u8] {
+        &self.mmap[self.frame_tables_data.clone()]
+    }
+
     /// Returns the contents of the `ELF_WASMTIME_INFO` section, or an empty
     /// slice if it wasn't found.
     #[inline]
@@ -305,13 +329,6 @@ impl CodeMemory {
         //   both the actual unwinding tables as well as the validity of the
         //   pointers we pass in itself.
         unsafe {
-            // First, if necessary, apply relocations. This can happen for
-            // things like libcalls which happen late in the lowering process
-            // that don't go through the Wasm-based libcalls layer that's
-            // indirected through the `VMContext`. Note that most modules won't
-            // have relocations, so this typically doesn't do anything.
-            self.apply_relocations()?;
-
             // Next freeze the contents of this image by making all of the
             // memory readonly. Nothing after this point should ever be modified
             // so commit everything. For a compiled-in-memory image this will
@@ -324,33 +341,20 @@ impl CodeMemory {
             // we aren't able to make it readonly, but this is just a
             // defense-in-depth measure and isn't required for correctness.
             #[cfg(has_virtual_memory)]
-            self.mmap.make_readonly(0..self.mmap.len())?;
+            if self.mmap.supports_virtual_memory() {
+                self.mmap.make_readonly(0..self.mmap.len())?;
+            }
 
             // Switch the executable portion from readonly to read/execute.
             if self.needs_executable {
                 if !self.custom_publish()? {
-                    #[cfg(has_virtual_memory)]
-                    {
-                        let text = self.text();
-
-                        use wasmtime_jit_icache_coherence as icache_coherence;
-
-                        // Clear the newly allocated code from cache if the processor requires it
-                        //
-                        // Do this before marking the memory as R+X, technically we should be able to do it after
-                        // but there are some CPU's that have had errata about doing this with read only memory.
-                        icache_coherence::clear_cache(text.as_ptr().cast(), text.len())
-                            .expect("Failed cache clear");
-
-                        self.mmap
-                            .make_executable(self.text.clone(), self.enable_branch_protection)
-                            .context("unable to make memory executable")?;
-
-                        // Flush any in-flight instructions from the pipeline
-                        icache_coherence::pipeline_flush_mt().expect("Failed pipeline flush");
+                    if !self.mmap.supports_virtual_memory() {
+                        bail!("this target requires virtual memory to be enabled");
                     }
-                    #[cfg(not(has_virtual_memory))]
-                    bail!("this target requires virtual memory to be enabled");
+                    #[cfg(has_virtual_memory)]
+                    self.mmap
+                        .make_executable(self.text.clone(), self.enable_branch_protection)
+                        .context("unable to make memory executable")?;
                 }
             }
 
@@ -390,39 +394,6 @@ impl CodeMemory {
         }
     }
 
-    unsafe fn apply_relocations(&mut self) -> Result<()> {
-        if self.relocations.is_empty() {
-            return Ok(());
-        }
-
-        for (offset, libcall) in self.relocations.iter() {
-            let offset = self.text.start + offset;
-            let libcall = match libcall {
-                obj::LibCall::FloorF32 => libcalls::relocs::floorf32 as usize,
-                obj::LibCall::FloorF64 => libcalls::relocs::floorf64 as usize,
-                obj::LibCall::NearestF32 => libcalls::relocs::nearestf32 as usize,
-                obj::LibCall::NearestF64 => libcalls::relocs::nearestf64 as usize,
-                obj::LibCall::CeilF32 => libcalls::relocs::ceilf32 as usize,
-                obj::LibCall::CeilF64 => libcalls::relocs::ceilf64 as usize,
-                obj::LibCall::TruncF32 => libcalls::relocs::truncf32 as usize,
-                obj::LibCall::TruncF64 => libcalls::relocs::truncf64 as usize,
-                obj::LibCall::FmaF32 => libcalls::relocs::fmaf32 as usize,
-                obj::LibCall::FmaF64 => libcalls::relocs::fmaf64 as usize,
-                #[cfg(target_arch = "x86_64")]
-                obj::LibCall::X86Pshufb => libcalls::relocs::x86_pshufb as usize,
-                #[cfg(not(target_arch = "x86_64"))]
-                obj::LibCall::X86Pshufb => unreachable!(),
-            };
-            self.mmap
-                .as_mut_slice()
-                .as_mut_ptr()
-                .add(offset)
-                .cast::<usize>()
-                .write_unaligned(libcall);
-        }
-        Ok(())
-    }
-
     unsafe fn register_unwind_info(&mut self) -> Result<()> {
         if self.unwind.len() == 0 {
             return Ok(());
@@ -431,12 +402,14 @@ impl CodeMemory {
         {
             let text = self.text();
             let unwind_info = &self.mmap[self.unwind.clone()];
-            let registration = crate::runtime::vm::UnwindRegistration::new(
-                text.as_ptr(),
-                unwind_info.as_ptr(),
-                unwind_info.len(),
-            )
-            .context("failed to create unwind info registration")?;
+            let registration = unsafe {
+                crate::runtime::vm::UnwindRegistration::new(
+                    text.as_ptr(),
+                    unwind_info.as_ptr(),
+                    unwind_info.len(),
+                )
+                .context("failed to create unwind info registration")?
+            };
             self.unwind_registration = Some(registration);
             return Ok(());
         }
@@ -457,7 +430,7 @@ impl CodeMemory {
         // and anything else necessary that is done in "create_gdbjit_image" right now.
         let image = self.mmap().to_vec();
         let text: &[u8] = self.text();
-        let bytes = crate::debug::create_gdbjit_image(image, (text.as_ptr(), text.len()))?;
+        let bytes = crate::native_debug::create_gdbjit_image(image, (text.as_ptr(), text.len()))?;
         let reg = crate::runtime::vm::GdbJitImageRegistration::register(bytes);
         self.debug_registration = Some(reg);
         Ok(())
@@ -468,6 +441,23 @@ impl CodeMemory {
     pub fn lookup_trap_code(&self, text_offset: usize) -> Option<Trap> {
         lookup_trap_code(self.trap_data(), text_offset)
     }
+
+    /// Get the raw address range of this CodeMemory.
+    pub(crate) fn raw_addr_range(&self) -> Range<usize> {
+        let start = self.text().as_ptr().addr();
+        let end = start + self.text().len();
+        start..end
+    }
+
+    /// Create a "deep clone": a separate CodeMemory for the same code
+    /// that can be patched or mutated independently. Also returns a
+    /// "metadata and location" handle that can be registered with the
+    /// global module registry and used for trap metadata lookups.
+    #[cfg(feature = "debug")]
+    pub(crate) fn deep_clone(self: &Arc<Self>, engine: &Engine) -> Result<CodeMemory> {
+        let mmap = self.mmap.deep_clone()?;
+        Self::new(engine, mmap)
+    }
 }
 
 /// Returns the range of `inner` within `outer`, such that `outer[range]` is the
