@@ -1,30 +1,26 @@
+use crate::component::InstanceExportLookup;
 use crate::component::matching::InstanceType;
 use crate::component::types;
-use crate::component::InstanceExportLookup;
 use crate::prelude::*;
-use crate::runtime::vm::component::ComponentRuntimeInfo;
 #[cfg(feature = "std")]
 use crate::runtime::vm::open_file_for_mmap;
-use crate::runtime::vm::{
-    CompiledModuleId, VMArrayCallFunction, VMFuncRef, VMFunctionBody, VMWasmCallFunction,
-};
+use crate::runtime::vm::{CompiledModuleId, VMArrayCallFunction, VMFuncRef, VMWasmCallFunction};
 use crate::{
-    code::CodeObject, code_memory::CodeMemory, type_registry::TypeCollection, Engine, Module,
-    ResourcesRequired,
+    Engine, Module, ResourcesRequired, code::EngineCode, code_memory::CodeMemory,
+    type_registry::TypeCollection,
 };
 use crate::{FuncType, ValType};
 use alloc::sync::Arc;
-use core::any::Any;
 use core::ops::Range;
 use core::ptr::NonNull;
 #[cfg(feature = "std")]
 use std::path::Path;
 use wasmtime_environ::component::{
-    AllCallFunc, CompiledComponentInfo, ComponentArtifacts, ComponentTypes, Export, ExportIndex,
-    GlobalInitializer, InstantiateModule, NameMapNoIntern, StaticModuleIndex, TrampolineIndex,
-    TypeComponentIndex, TypeDef, VMComponentOffsets,
+    CompiledComponentInfo, ComponentArtifacts, ComponentTypes, CoreDef, Export, ExportIndex,
+    GlobalInitializer, InstantiateModule, NameMapNoIntern, OptionsIndex, StaticModuleIndex,
+    TrampolineIndex, TypeComponentIndex, TypeFuncIndex, UnsafeIntrinsic, VMComponentOffsets,
 };
-use wasmtime_environ::TypeTrace;
+use wasmtime_environ::{Abi, CompiledFunctionsTable, FuncKey, TypeTrace};
 use wasmtime_environ::{FunctionLoc, HostPtr, ObjectKind, PrimaryMap};
 
 /// A compiled WebAssembly Component.
@@ -85,15 +81,19 @@ struct ComponentInner {
     ///
     /// Note that the `Arc` here is used to share this allocation with internal
     /// modules.
-    code: Arc<CodeObject>,
+    code: Arc<EngineCode>,
 
     /// Metadata produced during compilation.
     info: CompiledComponentInfo,
 
+    /// The index of compiled functions and their locations in the text section
+    /// for this component.
+    index: Arc<CompiledFunctionsTable>,
+
     /// A cached handle to the `wasmtime::FuncType` for the canonical ABI's
     /// `realloc`, to avoid the need to look up types in the registry and take
     /// locks when calling `realloc` via `TypedFunc::call_raw`.
-    realloc_func_type: Arc<dyn Any + Send + Sync>,
+    realloc_func_type: Arc<FuncType>,
 }
 
 pub(crate) struct AllCallFuncPointers {
@@ -227,7 +227,9 @@ impl Component {
     /// lives for as long as the module and is nevery externally modified for
     /// the lifetime of the deserialized module.
     pub unsafe fn deserialize_raw(engine: &Engine, memory: NonNull<[u8]>) -> Result<Component> {
-        let code = engine.load_code_raw(memory, ObjectKind::Component)?;
+        // SAFETY: the contract required by `load_code_raw` is the same as this
+        // function.
+        let code = unsafe { engine.load_code_raw(memory, ObjectKind::Component)? };
         Component::from_parts(engine, code, None)
     }
 
@@ -402,12 +404,14 @@ impl Component {
         let ComponentArtifacts {
             ty,
             info,
+            table: index,
             mut types,
             mut static_modules,
         } = match artifacts {
             Some(artifacts) => artifacts,
             None => postcard::from_bytes(code_memory.wasmtime_info())?,
         };
+        let index = Arc::new(index);
 
         // Validate that the component can be used with the current instance
         // allocator.
@@ -426,24 +430,26 @@ impl Component {
         );
         types.canonicalize_for_runtime_usage(&mut |idx| signatures.shared_type(idx).unwrap());
 
-        // Assemble the `CodeObject` artifact which is shared by all core wasm
+        // Assemble the `EngineCode` artifact which is shared by all core wasm
         // modules as well as the final component.
         let types = Arc::new(types);
-        let code = Arc::new(CodeObject::new(code_memory, signatures, types.into()));
+        let code = Arc::new(EngineCode::new(code_memory, signatures, types.into()));
 
         // Convert all information about static core wasm modules into actual
         // `Module` instances by converting each `CompiledModuleInfo`, the
         // `types` type information, and the code memory to a runtime object.
         let static_modules = static_modules
             .into_iter()
-            .map(|(_, info)| Module::from_parts_raw(engine, code.clone(), info, false))
+            .map(|(_, info)| {
+                Module::from_parts_raw(engine, code.clone(), info, index.clone(), false)
+            })
             .collect::<Result<_>>()?;
 
         let realloc_func_type = Arc::new(FuncType::new(
             engine,
             [ValType::I32, ValType::I32, ValType::I32, ValType::I32],
             [ValType::I32],
-        )) as _;
+        ));
 
         Ok(Component {
             inner: Arc::new(ComponentInner {
@@ -453,6 +459,7 @@ impl Component {
                 static_modules,
                 code,
                 info,
+                index,
                 realloc_func_type,
             }),
         })
@@ -470,42 +477,79 @@ impl Component {
         &self.inner.static_modules[idx]
     }
 
-    #[cfg_attr(not(feature = "profiling"), allow(dead_code))]
+    #[cfg(feature = "profiling")]
     pub(crate) fn static_modules(&self) -> impl Iterator<Item = &Module> {
         self.inner.static_modules.values()
     }
 
     #[inline]
     pub(crate) fn types(&self) -> &Arc<ComponentTypes> {
-        self.inner.component_types()
+        match self.inner.code.types() {
+            crate::code::Types::Component(types) => types,
+            // The only creator of a `Component` is itself which uses the other
+            // variant, so this shouldn't be possible.
+            crate::code::Types::Module(_) => unreachable!(),
+        }
     }
 
     pub(crate) fn signatures(&self) -> &TypeCollection {
         self.inner.code.signatures()
     }
 
-    pub(crate) fn text(&self) -> &[u8] {
-        self.inner.code.code_memory().text()
-    }
-
     pub(crate) fn trampoline_ptrs(&self, index: TrampolineIndex) -> AllCallFuncPointers {
-        let AllCallFunc {
+        let wasm_call = self
+            .store_invariant_func(FuncKey::ComponentTrampoline(Abi::Wasm, index))
+            .unwrap()
+            .cast();
+        let array_call = self
+            .store_invariant_func(FuncKey::ComponentTrampoline(Abi::Array, index))
+            .unwrap()
+            .cast();
+        AllCallFuncPointers {
             wasm_call,
             array_call,
-        } = &self.inner.info.trampolines[index];
-        AllCallFuncPointers {
-            wasm_call: self.func(wasm_call).cast(),
-            array_call: self.func(array_call).cast(),
         }
     }
 
-    fn func(&self, loc: &FunctionLoc) -> NonNull<VMFunctionBody> {
-        let text = self.text();
-        let trampoline = &text[loc.start as usize..][..loc.length as usize];
-        NonNull::new(trampoline.as_ptr() as *mut VMFunctionBody).unwrap()
+    pub(crate) fn unsafe_intrinsic_ptrs(
+        &self,
+        intrinsic: UnsafeIntrinsic,
+    ) -> Option<AllCallFuncPointers> {
+        let wasm_call = self
+            .store_invariant_func(FuncKey::UnsafeIntrinsic(Abi::Wasm, intrinsic))?
+            .cast();
+        let array_call = self
+            .store_invariant_func(FuncKey::UnsafeIntrinsic(Abi::Array, intrinsic))?
+            .cast();
+        Some(AllCallFuncPointers {
+            wasm_call,
+            array_call,
+        })
     }
 
-    pub(crate) fn code_object(&self) -> &Arc<CodeObject> {
+    /// Look up a function in this component's text section by `FuncKey`.
+    ///
+    /// This supports only `FuncKey`s that do not invoke Wasm code,
+    /// i.e., code that is potentially Store-specific.
+    fn store_invariant_func(&self, key: FuncKey) -> Option<NonNull<u8>> {
+        assert!(key.is_store_invariant());
+        let loc = self.inner.index.func_loc(key)?;
+        Some(self.func_loc_to_pointer(loc))
+    }
+
+    /// Given a function location within this component's text section, get a
+    /// pointer to the function.
+    ///
+    /// This works only for Store-invariant functions.
+    ///
+    /// Panics on out-of-bounds function locations.
+    fn func_loc_to_pointer(&self, loc: &FunctionLoc) -> NonNull<u8> {
+        let text = self.engine_code().text();
+        let trampoline = &text[loc.start as usize..][..loc.length as usize];
+        NonNull::from(trampoline).cast()
+    }
+
+    pub(crate) fn engine_code(&self) -> &Arc<EngineCode> {
         &self.inner.code
     }
 
@@ -518,11 +562,7 @@ impl Component {
     /// [`Module::serialize`]: crate::Module::serialize
     /// [`Module`]: crate::Module
     pub fn serialize(&self) -> Result<Vec<u8>> {
-        Ok(self.code_object().code_memory().mmap().to_vec())
-    }
-
-    pub(crate) fn runtime_info(&self) -> Arc<dyn ComponentRuntimeInfo> {
-        self.inner.clone()
+        Ok(self.engine_code().image().to_vec())
     }
 
     /// Creates a new `VMFuncRef` with all fields filled out for the destructor
@@ -539,11 +579,9 @@ impl Component {
         // then this can't be called by the component, so it's ok to leave it
         // blank.
         let wasm_call = self
-            .inner
-            .info
-            .resource_drop_wasm_to_array_trampoline
-            .as_ref()
-            .map(|i| self.func(i).cast().into());
+            .store_invariant_func(FuncKey::ResourceDropTrampoline)
+            .map(|f| f.cast().into());
+
         VMFuncRef {
             wasm_call,
             ..*dtor.func_ref()
@@ -644,7 +682,10 @@ impl Component {
     /// For more information see
     /// [`Module::image_range`](crate::Module::image_range).
     pub fn image_range(&self) -> Range<*const u8> {
-        self.inner.code.code_memory().mmap().image_range()
+        let range = self.inner.code.text_range();
+        let start = range.start.raw() as *const u8;
+        let end = range.end.raw() as *const u8;
+        start..end
     }
 
     /// Force initialization of copy-on-write images to happen here-and-now
@@ -671,6 +712,37 @@ impl Component {
         Ok(())
     }
 
+    /// Looks up a specific export of this component by `name` optionally nested
+    /// within the `instance` provided.
+    ///
+    /// See related method [`Self::get_export`] for additional docs and
+    /// examples.
+    ///
+    /// This method is primarily used to acquire a [`ComponentExportIndex`]
+    /// which can be used with [`Instance`](crate::component::Instance) when
+    /// looking up exports. Export lookup with [`ComponentExportIndex`] can
+    /// skip string lookups at runtime and instead use a more efficient
+    /// index-based lookup.
+    ///
+    /// This method only returns the [`ComponentExportIndex`]. If you need the
+    /// corresponding [`types::ComponentItem`], use the related function
+    /// [`Self::get_export`].
+    ///
+    ///
+    /// [`Instance`](crate::component::Instance) has a corresponding method
+    /// [`Instance::get_export_index`](crate::component::Instance::get_export_index).
+    pub fn get_export_index(
+        &self,
+        instance: Option<&ComponentExportIndex>,
+        name: &str,
+    ) -> Option<ComponentExportIndex> {
+        let index = self.lookup_export_index(instance, name)?;
+        Some(ComponentExportIndex {
+            id: self.inner.id,
+            index,
+        })
+    }
+
     /// Looks up a specific export of this component by `name` optionally nested
     /// within the `instance` provided.
     ///
@@ -697,6 +769,14 @@ impl Component {
     /// with [`Instance::get_func`](crate::component::Instance::get_func) for
     /// example.
     ///
+    /// The returned [`types::ComponentItem`] is more expensive to calculate
+    /// than the [`ComponentExportIndex`]. If you only consume the
+    /// [`ComponentExportIndex`], use the related method
+    /// [`Self::get_export_index`] instead.
+    ///
+    /// [`Instance`](crate::component::Instance) has a corresponding method
+    /// [`Instance::get_export`](crate::component::Instance::get_export).
+    ///
     /// # Examples
     ///
     /// ```
@@ -721,7 +801,7 @@ impl Component {
     /// )?;
     ///
     /// // Perform a lookup of the function "f" before instantiaton.
-    /// let (ty, export) = component.export_index(None, "f").unwrap();
+    /// let (ty, export) = component.get_export(None, "f").unwrap();
     /// assert!(matches!(ty, ComponentItem::ComponentFunc(_)));
     ///
     /// // After instantiation use `export` to lookup the function in question
@@ -733,23 +813,19 @@ impl Component {
     /// # Ok(())
     /// # }
     /// ```
-    pub fn export_index(
+    pub fn get_export(
         &self,
         instance: Option<&ComponentExportIndex>,
         name: &str,
     ) -> Option<(types::ComponentItem, ComponentExportIndex)> {
         let info = self.env_component();
         let index = self.lookup_export_index(instance, name)?;
-        let ty = match info.export_items[index] {
-            Export::Instance { ty, .. } => TypeDef::ComponentInstance(ty),
-            Export::LiftedFunction { ty, .. } => TypeDef::ComponentFunc(ty),
-            Export::ModuleStatic { ty, .. } | Export::ModuleImport { ty, .. } => {
-                TypeDef::Module(ty)
-            }
-            Export::Type(ty) => ty,
-        };
         let item = self.with_uninstantiated_instance_type(|instance| {
-            types::ComponentItem::from(&self.inner.engine, &ty, instance)
+            types::ComponentItem::from_export(
+                &self.inner.engine,
+                &info.export_items[index],
+                instance,
+            )
         });
         Some((
             item,
@@ -789,11 +865,31 @@ impl Component {
     pub fn engine(&self) -> &Engine {
         &self.inner.engine
     }
+
+    pub(crate) fn realloc_func_ty(&self) -> &Arc<FuncType> {
+        &self.inner.realloc_func_type
+    }
+
+    /// Returns the `Export::LiftedFunction` metadata associated with `export`.
+    ///
+    /// # Panics
+    ///
+    /// Panics if `export` is out of bounds or if it isn't a `LiftedFunction`.
+    pub(crate) fn export_lifted_function(
+        &self,
+        export: ExportIndex,
+    ) -> (TypeFuncIndex, &CoreDef, OptionsIndex) {
+        let component = self.env_component();
+        match &component.export_items[export] {
+            Export::LiftedFunction { ty, func, options } => (*ty, func, *options),
+            _ => unreachable!(),
+        }
+    }
 }
 
 /// A value which represents a known export of a component.
 ///
-/// This is the return value of [`Component::export_index`] and implements the
+/// This is the return value of [`Component::get_export`] and implements the
 /// [`InstanceExportLookup`] trait to work with lookups like
 /// [`Instance::get_func`](crate::component::Instance::get_func).
 #[derive(Copy, Clone, Debug, Hash, Eq, PartialEq)]
@@ -812,25 +908,6 @@ impl InstanceExportLookup for ComponentExportIndex {
     }
 }
 
-impl ComponentRuntimeInfo for ComponentInner {
-    fn component(&self) -> &wasmtime_environ::component::Component {
-        &self.info.component
-    }
-
-    fn component_types(&self) -> &Arc<ComponentTypes> {
-        match self.code.types() {
-            crate::code::Types::Component(types) => types,
-            // The only creator of a `Component` is itself which uses the other
-            // variant, so this shouldn't be possible.
-            crate::code::Types::Module(_) => unreachable!(),
-        }
-    }
-
-    fn realloc_func_type(&self) -> &Arc<dyn Any + Send + Sync> {
-        &self.realloc_func_type
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use crate::component::Component;
