@@ -1,7 +1,7 @@
 use crate::component::concurrent;
 #[cfg(feature = "component-model-async")]
 use crate::component::concurrent::{Accessor, Status};
-use crate::component::func::{LiftContext, LowerContext, Options};
+use crate::component::func::{LiftContext, LowerContext};
 use crate::component::matching::InstanceType;
 use crate::component::storage::slice_to_storage_mut;
 use crate::component::types::ComponentFunc;
@@ -41,18 +41,37 @@ enum HostResult<T> {
     Future(Pin<Box<dyn Future<Output = Result<T>> + Send>>),
 }
 
+trait FunctionStyle {
+    const ASYNC: bool;
+}
+
+struct SyncStyle;
+
+impl FunctionStyle for SyncStyle {
+    const ASYNC: bool = false;
+}
+
+#[cfg(feature = "component-model-async")]
+struct AsyncStyle;
+
+#[cfg(feature = "component-model-async")]
+impl FunctionStyle for AsyncStyle {
+    const ASYNC: bool = true;
+}
+
 impl HostFunc {
-    fn from_canonical<T, F, P, R>(func: F) -> Arc<HostFunc>
+    fn from_canonical<T, F, P, R, S>(func: F) -> Arc<HostFunc>
     where
         F: Fn(StoreContextMut<'_, T>, P) -> HostResult<R> + Send + Sync + 'static,
         P: ComponentNamedList + Lift + 'static,
         R: ComponentNamedList + Lower + 'static,
         T: 'static,
+        S: FunctionStyle + 'static,
     {
-        let entrypoint = Self::entrypoint::<T, F, P, R>;
+        let entrypoint = Self::entrypoint::<T, F, P, R, S>;
         Arc::new(HostFunc {
             entrypoint,
-            typecheck: Box::new(typecheck::<P, R>),
+            typecheck: Box::new(typecheck::<P, R, S>),
             func: Box::new(func),
         })
     }
@@ -64,7 +83,7 @@ impl HostFunc {
         P: ComponentNamedList + Lift + 'static,
         R: ComponentNamedList + Lower + 'static,
     {
-        Self::from_canonical::<T, _, _, _>(move |store, params| {
+        Self::from_canonical::<T, _, _, _, SyncStyle>(move |store, params| {
             HostResult::Done(func(store, params))
         })
     }
@@ -81,7 +100,7 @@ impl HostFunc {
         R: ComponentNamedList + Lower + 'static,
     {
         let func = Arc::new(func);
-        Self::from_canonical::<T, _, _, _>(move |store, params| {
+        Self::from_canonical::<T, _, _, _, AsyncStyle>(move |store, params| {
             let func = func.clone();
             HostResult::Future(Box::pin(
                 store.wrap_call(move |accessor| func(accessor, params)),
@@ -89,7 +108,7 @@ impl HostFunc {
         })
     }
 
-    extern "C" fn entrypoint<T, F, P, R>(
+    extern "C" fn entrypoint<T, F, P, R, S>(
         cx: NonNull<VMOpaqueContext>,
         data: NonNull<u8>,
         ty: u32,
@@ -102,11 +121,12 @@ impl HostFunc {
         P: ComponentNamedList + Lift,
         R: ComponentNamedList + Lower + 'static,
         T: 'static,
+        S: FunctionStyle,
     {
         let data = SendSyncPtr::new(NonNull::new(data.as_ptr() as *mut F).unwrap());
         unsafe {
             call_host_and_handle_result::<T>(cx, |store, instance| {
-                call_host(
+                call_host::<_, _, _, _, S>(
                     store,
                     instance,
                     TypeFuncIndex::from_u32(ty),
@@ -118,7 +138,7 @@ impl HostFunc {
         }
     }
 
-    fn new_dynamic_canonical<T, F>(func: F) -> Arc<HostFunc>
+    fn new_dynamic_canonical<T, F, S>(func: F) -> Arc<HostFunc>
     where
         F: Fn(
                 StoreContextMut<'_, T>,
@@ -130,13 +150,21 @@ impl HostFunc {
             + Sync
             + 'static,
         T: 'static,
+        S: FunctionStyle,
     {
         Arc::new(HostFunc {
-            entrypoint: dynamic_entrypoint::<T, F>,
-            // This function performs dynamic type checks and subsequently does
-            // not need to perform up-front type checks. Instead everything is
-            // dynamically managed at runtime.
-            typecheck: Box::new(move |_expected_index, _expected_types| Ok(())),
+            entrypoint: dynamic_entrypoint::<T, F, S>,
+            // This function performs dynamic type checks on its parameters and
+            // results and subsequently does not need to perform up-front type
+            // checks. However, we _do_ verify async-ness here.
+            typecheck: Box::new(move |ty, types| {
+                let ty = &types.types[ty];
+                if S::ASYNC != ty.async_ {
+                    bail!("type mismatch with async");
+                }
+
+                Ok(())
+            }),
             func: Box::new(func),
         })
     }
@@ -148,7 +176,7 @@ impl HostFunc {
             + Sync
             + 'static,
     {
-        Self::new_dynamic_canonical::<T, _>(
+        Self::new_dynamic_canonical::<T, _, SyncStyle>(
             move |store, ty, mut params_and_results, result_start| {
                 let (params, results) = params_and_results.split_at_mut(result_start);
                 let result = func(store, ty, params, results).map(move |()| params_and_results);
@@ -172,7 +200,7 @@ impl HostFunc {
             + 'static,
     {
         let func = Arc::new(func);
-        Self::new_dynamic_canonical::<T, _>(
+        Self::new_dynamic_canonical::<T, _, AsyncStyle>(
             move |store, ty, mut params_and_results, result_start| {
                 let func = func.clone();
                 Box::pin(store.wrap_call(move |accessor| {
@@ -199,12 +227,16 @@ impl HostFunc {
     }
 }
 
-fn typecheck<P, R>(ty: TypeFuncIndex, types: &InstanceType<'_>) -> Result<()>
+fn typecheck<P, R, S>(ty: TypeFuncIndex, types: &InstanceType<'_>) -> Result<()>
 where
     P: ComponentNamedList + Lift,
     R: ComponentNamedList + Lower,
+    S: FunctionStyle,
 {
     let ty = &types.types[ty];
+    if S::ASYNC != ty.async_ {
+        bail!("type mismatch with async");
+    }
     P::typecheck(&InterfaceType::Tuple(ty.params), types)
         .context("type mismatch with parameters")?;
     R::typecheck(&InterfaceType::Tuple(ty.results), types).context("type mismatch with results")?;
@@ -225,6 +257,7 @@ where
 /// * `Return` - the result of the host function
 /// * `F` - the `closure` to actually receive the `Params` and return the
 ///   `Return`
+/// * `S` - the expected `FunctionStyle`
 ///
 /// It's expected that `F` will "un-tuple" the arguments to pass to a host
 /// closure.
@@ -232,11 +265,11 @@ where
 /// This function is in general `unsafe` as the validity of all the parameters
 /// must be upheld. Generally that's done by ensuring this is only called from
 /// the select few places it's intended to be called from.
-unsafe fn call_host<T, Params, Return, F>(
-    mut store: StoreContextMut<'_, T>,
+unsafe fn call_host<T, Params, Return, F, S>(
+    store: StoreContextMut<'_, T>,
     instance: Instance,
     ty: TypeFuncIndex,
-    options_idx: OptionsIndex,
+    options: OptionsIndex,
     storage: &mut [MaybeUninit<ValRaw>],
     closure: F,
 ) -> Result<()>
@@ -244,11 +277,13 @@ where
     F: Fn(StoreContextMut<'_, T>, Params) -> HostResult<Return> + Send + Sync + 'static,
     Params: Lift,
     Return: Lower + 'static,
+    S: FunctionStyle,
 {
-    let options = Options::new_index(store.0, instance, options_idx);
+    let (component, store) = instance.component_and_store_mut(store.0);
+    let mut store = StoreContextMut(store);
     let vminstance = instance.id().get(store.0);
-    let opts = &vminstance.component().env_component().options[options_idx];
-    let async_ = opts.async_;
+    let opts = &vminstance.component().env_component().options[options];
+    let async_lower = opts.async_;
     let caller_instance = opts.instance;
     let mut flags = vminstance.instance_flags(caller_instance);
 
@@ -259,27 +294,26 @@ where
         return Err(anyhow!(crate::Trap::CannotLeaveComponent));
     }
 
-    let types = vminstance.component().types().clone();
+    let types = component.types();
     let ty = &types[ty];
     let param_tys = InterfaceType::Tuple(ty.params);
     let result_tys = InterfaceType::Tuple(ty.results);
 
-    if async_ {
+    if async_lower {
         #[cfg(feature = "component-model-async")]
         {
             let mut storage = unsafe { Storage::<'_, Params, u32>::new_async::<Return>(storage) };
 
             // Lift the parameters, either from flat storage or from linear
             // memory.
-            let lift = &mut LiftContext::new(store.0.store_opaque_mut(), &options, instance);
+            let lift = &mut LiftContext::new(store.0.store_opaque_mut(), options, instance);
             lift.enter_call();
             let params = storage.lift_params(lift, param_tys)?;
 
             // Load the return pointer, if present.
             let retptr = match storage.async_retptr() {
                 Some(ptr) => {
-                    let mut lower =
-                        LowerContext::new(store.as_context_mut(), &options, &types, instance);
+                    let mut lower = LowerContext::new(store.as_context_mut(), options, instance);
                     validate_inbounds::<Return>(lower.as_slice_mut(), ptr)?
                 }
                 // If there's no return pointer then `Return` should have an
@@ -295,12 +329,11 @@ where
             let host_result = closure(store.as_context_mut(), params);
 
             let mut lower_result = {
-                let types = types.clone();
                 move |store: StoreContextMut<T>, ret: Return| {
                     unsafe {
                         flags.set_may_leave(false);
                     }
-                    let mut lower = LowerContext::new(store, &options, &types, instance);
+                    let mut lower = LowerContext::new(store, options, instance);
                     ret.linear_lower_to_memory(&mut lower, result_tys, retptr)?;
                     unsafe {
                         flags.set_may_leave(true);
@@ -329,7 +362,7 @@ where
                 Status::Returned.pack(None)
             };
 
-            let mut lower = LowerContext::new(store, &options, &types, instance);
+            let mut lower = LowerContext::new(store, options, instance);
             storage.lower_results(&mut lower, InterfaceType::U32, status)?;
         }
         #[cfg(not(feature = "component-model-async"))]
@@ -341,8 +374,15 @@ where
             );
         }
     } else {
+        if S::ASYNC {
+            // The caller has synchronously lowered an async function, meaning
+            // the caller can only call it from an async task (i.e. a task
+            // created via a call to an async export).  Otherwise, we'll trap.
+            concurrent::check_blocking(store.0)?;
+        }
+
         let mut storage = unsafe { Storage::<'_, Params, Return>::new_sync(storage) };
-        let mut lift = LiftContext::new(store.0.store_opaque_mut(), &options, instance);
+        let mut lift = LiftContext::new(store.0.store_opaque_mut(), options, instance);
         lift.enter_call();
         let params = storage.lift_params(&mut lift, param_tys)?;
 
@@ -357,7 +397,7 @@ where
         unsafe {
             flags.set_may_leave(false);
         }
-        let mut lower = LowerContext::new(store, &options, &types, instance);
+        let mut lower = LowerContext::new(store, options, instance);
         storage.lower_results(&mut lower, result_tys, ret)?;
         unsafe {
             flags.set_may_leave(true);
@@ -699,11 +739,11 @@ where
     }
 }
 
-unsafe fn call_host_dynamic<T, F>(
-    mut store: StoreContextMut<'_, T>,
+unsafe fn call_host_dynamic<T, F, S>(
+    store: StoreContextMut<'_, T>,
     instance: Instance,
     ty: TypeFuncIndex,
-    options_idx: OptionsIndex,
+    options: OptionsIndex,
     storage: &mut [MaybeUninit<ValRaw>],
     closure: F,
 ) -> Result<()>
@@ -718,11 +758,13 @@ where
         + Sync
         + 'static,
     T: 'static,
+    S: FunctionStyle,
 {
-    let options = Options::new_index(store.0, instance, options_idx);
+    let (component, store) = instance.component_and_store_mut(store.0);
+    let mut store = StoreContextMut(store);
     let vminstance = instance.id().get(store.0);
-    let opts = &vminstance.component().env_component().options[options_idx];
-    let async_ = opts.async_;
+    let opts = &component.env_component().options[options];
+    let async_lower = opts.async_;
     let caller_instance = opts.instance;
     let mut flags = vminstance.instance_flags(caller_instance);
 
@@ -733,15 +775,15 @@ where
         return Err(anyhow!(crate::Trap::CannotLeaveComponent));
     }
 
-    let types = instance.id().get(store.0).component().types().clone();
+    let types = component.types();
     let func_ty = &types[ty];
     let param_tys = &types[func_ty.params];
     let result_tys = &types[func_ty.results];
 
     let mut params_and_results = Vec::new();
-    let mut lift = &mut LiftContext::new(store.0.store_opaque_mut(), &options, instance);
+    let mut lift = &mut LiftContext::new(store.0.store_opaque_mut(), options, instance);
     lift.enter_call();
-    let max_flat = if async_ {
+    let max_flat = if async_lower {
         MAX_FLAT_ASYNC_PARAMS
     } else {
         MAX_FLAT_PARAMS
@@ -751,7 +793,7 @@ where
     let ret_index = unsafe {
         dynamic_params_load(
             &mut lift,
-            &types,
+            types,
             storage,
             param_tys,
             &mut params_and_results,
@@ -763,36 +805,33 @@ where
         params_and_results.push(Val::Bool(false));
     }
 
-    if async_ {
+    if async_lower {
         #[cfg(feature = "component-model-async")]
         {
             let retptr = if result_tys.types.len() == 0 {
                 0
             } else {
                 let retptr = unsafe { storage[ret_index].assume_init() };
-                let mut lower =
-                    LowerContext::new(store.as_context_mut(), &options, &types, instance);
+                let mut lower = LowerContext::new(store.as_context_mut(), options, instance);
                 validate_inbounds_dynamic(&result_tys.abi, lower.as_slice_mut(), &retptr)?
             };
 
             let future = closure(store.as_context_mut(), ty, params_and_results, result_start);
 
             let task = instance.first_poll(store, future, caller_instance, {
-                let types = types.clone();
                 let result_tys = func_ty.results;
                 move |store: StoreContextMut<T>, result_vals: Vec<Val>| {
-                    let result_tys = &types[result_tys];
-                    let result_vals = &result_vals[result_start..];
-                    assert_eq!(result_vals.len(), result_tys.types.len());
-
                     unsafe {
                         flags.set_may_leave(false);
                     }
 
-                    let mut lower = LowerContext::new(store, &options, &types, instance);
+                    let mut lower = LowerContext::new(store, options, instance);
+                    let result_tys = &lower.types[result_tys];
+                    let result_vals = &result_vals[result_start..];
+                    assert_eq!(result_vals.len(), result_tys.types.len());
                     let mut ptr = retptr;
                     for (val, ty) in result_vals.iter().zip(result_tys.types.iter()) {
-                        let offset = types.canonical_abi(ty).next_field32_size(&mut ptr);
+                        let offset = lower.types.canonical_abi(ty).next_field32_size(&mut ptr);
                         val.store(&mut lower, *ty, offset)?;
                     }
 
@@ -822,6 +861,13 @@ where
             );
         }
     } else {
+        if S::ASYNC {
+            // The caller has synchronously lowered an async function, meaning
+            // the caller can only call it from an async task (i.e. a task
+            // created via a call to an async export).  Otherwise, we'll trap.
+            concurrent::check_blocking(store.0)?;
+        }
+
         let future = closure(store.as_context_mut(), ty, params_and_results, result_start);
         let result_vals = concurrent::poll_and_block(store.0, future, caller_instance)?;
         let result_vals = &result_vals[result_start..];
@@ -830,7 +876,7 @@ where
             flags.set_may_leave(false);
         }
 
-        let mut cx = LowerContext::new(store, &options, &types, instance);
+        let mut cx = LowerContext::new(store, options, instance);
         if let Some(cnt) = result_tys.abi.flat_count(MAX_FLAT_RESULTS) {
             let mut dst = storage[..cnt].iter_mut();
             for (val, ty) in result_vals.iter().zip(result_tys.types.iter()) {
@@ -916,7 +962,7 @@ pub(crate) fn validate_inbounds_dynamic(
     Ok(ptr)
 }
 
-extern "C" fn dynamic_entrypoint<T, F>(
+extern "C" fn dynamic_entrypoint<T, F, S>(
     cx: NonNull<VMOpaqueContext>,
     data: NonNull<u8>,
     ty: u32,
@@ -935,11 +981,12 @@ where
         + Sync
         + 'static,
     T: 'static,
+    S: FunctionStyle,
 {
     let data = SendSyncPtr::new(NonNull::new(data.as_ptr() as *mut F).unwrap());
     unsafe {
         call_host_and_handle_result(cx, |store, instance| {
-            call_host_dynamic::<T, _>(
+            call_host_dynamic::<T, _, S>(
                 store,
                 instance,
                 TypeFuncIndex::from_u32(ty),
