@@ -2,16 +2,16 @@ use crate::func::HostFunc;
 use crate::hash_map::{Entry, HashMap};
 use crate::instance::InstancePre;
 use crate::store::StoreOpaque;
-use crate::{prelude::*, IntoFunc};
 use crate::{
     AsContext, AsContextMut, Caller, Engine, Extern, ExternType, Func, FuncType, ImportType,
     Instance, Module, StoreContextMut, Val, ValRaw,
 };
+use crate::{IntoFunc, prelude::*};
 use alloc::sync::Arc;
 use core::fmt::{self, Debug};
-use core::marker;
 #[cfg(feature = "async")]
-use core::{future::Future, pin::Pin};
+use core::future::Future;
+use core::marker;
 use log::warn;
 
 /// Structure used to link wasm modules/instances together.
@@ -240,7 +240,10 @@ impl<T> Linker<T> {
     /// # Ok(())
     /// # }
     /// ```
-    pub fn define_unknown_imports_as_traps(&mut self, module: &Module) -> anyhow::Result<()> {
+    pub fn define_unknown_imports_as_traps(&mut self, module: &Module) -> anyhow::Result<()>
+    where
+        T: 'static,
+    {
         for import in module.imports() {
             if let Err(import_err) = self._get_by_import(&import) {
                 if let ExternType::Func(func_ty) = import_err.ty() {
@@ -276,14 +279,23 @@ impl<T> Linker<T> {
     /// ```
     pub fn define_unknown_imports_as_default_values(
         &mut self,
-        store: &mut impl AsContextMut<Data = T>,
+        mut store: impl AsContextMut<Data = T>,
         module: &Module,
-    ) -> anyhow::Result<()> {
+    ) -> anyhow::Result<()>
+    where
+        T: 'static,
+    {
+        let mut store = store.as_context_mut();
         for import in module.imports() {
             if let Err(import_err) = self._get_by_import(&import) {
                 let default_extern =
-                    import_err.ty().default_value(&mut *store).ok_or_else(|| {
-                        anyhow!("no default value exists for type `{:?}`", import_err.ty())
+                    import_err.ty().default_value(&mut store).with_context(|| {
+                        anyhow!(
+                            "no default value exists for `{}::{}` with type `{:?}`",
+                            import.module(),
+                            import.name(),
+                            import_err.ty(),
+                        )
                     })?;
                 self.define(
                     store.as_context(),
@@ -338,7 +350,10 @@ impl<T> Linker<T> {
         module: &str,
         name: &str,
         item: impl Into<Extern>,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         let store = store.as_context();
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::new(store.0, item.into()))?;
@@ -356,7 +371,10 @@ impl<T> Linker<T> {
         store: impl AsContext<Data = T>,
         name: &str,
         item: impl Into<Extern>,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         let store = store.as_context();
         let key = self.import_key(name, None);
         self.insert(key, Definition::new(store.0, item.into()))?;
@@ -377,7 +395,10 @@ impl<T> Linker<T> {
         name: &str,
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<()> + Send + Sync + 'static,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         assert!(ty.comes_from_same_engine(self.engine()));
         let func = HostFunc::new(&self.engine, ty, func);
         let key = self.import_key(module, Some(name));
@@ -393,15 +414,23 @@ impl<T> Linker<T> {
     ///
     /// Panics if the given function type is not associated with the same engine
     /// as this linker.
+    ///
+    /// # Safety
+    ///
+    /// See [`Func::new_unchecked`] for more safety information.
     pub unsafe fn func_new_unchecked(
         &mut self,
         module: &str,
         name: &str,
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &mut [ValRaw]) -> Result<()> + Send + Sync + 'static,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         assert!(ty.comes_from_same_engine(self.engine()));
-        let func = HostFunc::new_unchecked(&self.engine, ty, func);
+        // SAFETY: the contract of this function is the same as `new_unchecked`.
+        let func = unsafe { HostFunc::new_unchecked(&self.engine, ty, func) };
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
         Ok(self)
@@ -437,24 +466,19 @@ impl<T> Linker<T> {
             + Send
             + Sync
             + 'static,
+        T: 'static,
     {
         assert!(
             self.engine.config().async_support,
             "cannot use `func_new_async` without enabling async support in the config"
         );
         assert!(ty.comes_from_same_engine(self.engine()));
-        self.func_new(module, name, ty, move |mut caller, params, results| {
-            let async_cx = caller
-                .store
-                .as_context_mut()
-                .0
-                .async_cx()
-                .expect("Attempt to spawn new function on dying fiber");
-            let future = func(caller, params, results);
-            match unsafe { async_cx.block_on(Pin::from(future)) } {
-                Ok(Ok(())) => Ok(()),
-                Ok(Err(trap)) | Err(trap) => Err(trap),
-            }
+        self.func_new(module, name, ty, move |caller, params, results| {
+            let instance = caller.caller();
+            caller.store.with_blocking(|store, cx| {
+                let caller = Caller::new(store, instance);
+                cx.block_on(core::pin::Pin::from(func(caller, params, results)))
+            })?
         })
     }
 
@@ -519,7 +543,10 @@ impl<T> Linker<T> {
         module: &str,
         name: &str,
         func: impl IntoFunc<T, Params, Args>,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         let func = HostFunc::wrap(&self.engine, func);
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
@@ -539,27 +566,24 @@ impl<T> Linker<T> {
             + Send
             + Sync
             + 'static,
+        T: 'static,
     {
         assert!(
             self.engine.config().async_support,
             "cannot use `func_wrap_async` without enabling async support on the config",
         );
-        let func = HostFunc::wrap_inner(
-            &self.engine,
-            move |mut caller: Caller<'_, T>, args: Params| {
-                let async_cx = caller
-                    .store
-                    .as_context_mut()
-                    .0
-                    .async_cx()
-                    .expect("Attempt to start async function on dying fiber");
-                let future = func(caller, args);
-                match unsafe { async_cx.block_on(Pin::from(future)) } {
+        let func =
+            HostFunc::wrap_inner(&self.engine, move |caller: Caller<'_, T>, args: Params| {
+                let instance = caller.caller();
+                let result = caller.store.block_on(|store| {
+                    let caller = Caller::new(store, instance);
+                    func(caller, args).into()
+                });
+                match result {
                     Ok(ret) => ret.into_fallible(),
                     Err(e) => Args::fallible_from_error(e),
                 }
-            },
-        );
+            });
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
         Ok(self)
@@ -626,7 +650,10 @@ impl<T> Linker<T> {
         mut store: impl AsContextMut<Data = T>,
         module_name: &str,
         instance: Instance,
-    ) -> Result<&mut Self> {
+    ) -> Result<&mut Self>
+    where
+        T: 'static,
+    {
         let mut store = store.as_context_mut();
         let exports = instance
             .exports(&mut store)
@@ -815,7 +842,7 @@ impl<T> Linker<T> {
                 if let Some(export) = instance.get_export(&mut store, "_initialize") {
                     if let Extern::Func(func) = export {
                         func.typed::<(), ()>(&store)
-                            .and_then(|f| f.call(&mut store, ()).map_err(Into::into))
+                            .and_then(|f| f.call(&mut store, ()))
                             .context("calling the Reactor initialization function")?;
                     }
                 }
@@ -935,7 +962,9 @@ impl<T> Linker<T> {
             } else if export.name() == "__rtti_base" && export.ty().global().is_some() {
                 // Allow an exported "__rtti_base" memory for compatibility with
                 // AssemblyScript.
-                warn!("command module exporting '__rtti_base' is deprecated; pass `--runtime half` to the AssemblyScript compiler");
+                warn!(
+                    "command module exporting '__rtti_base' is deprecated; pass `--runtime half` to the AssemblyScript compiler"
+                );
             } else if !self.allow_unknown_exports {
                 bail!("command export '{}' is not a function", export.name());
             }
@@ -964,7 +993,7 @@ impl<T> Linker<T> {
         let dst = self.import_key(as_module, Some(as_name));
         match self.map.get(&src).cloned() {
             Some(item) => self.insert(dst, item)?,
-            None => bail!("no item named `{}::{}` defined", module, name),
+            None => bail!("no item named `{module}::{name}` defined"),
         }
         Ok(self)
     }
@@ -1007,7 +1036,7 @@ impl<T> Linker<T> {
                     Some(name) => format!("{module}::{name}"),
                     None => module.to_string(),
                 };
-                bail!("import of `{}` defined twice", desc)
+                bail!("import of `{desc}` defined twice")
             }
             Entry::Occupied(mut o) => {
                 o.insert(item);
@@ -1093,7 +1122,10 @@ impl<T> Linker<T> {
         &self,
         mut store: impl AsContextMut<Data = T>,
         module: &Module,
-    ) -> Result<Instance> {
+    ) -> Result<Instance>
+    where
+        T: 'static,
+    {
         self._instantiate_pre(module, Some(store.as_context_mut().0))?
             .instantiate(store)
     }
@@ -1107,7 +1139,7 @@ impl<T> Linker<T> {
         module: &Module,
     ) -> Result<Instance>
     where
-        T: Send,
+        T: Send + 'static,
     {
         self._instantiate_pre(module, Some(store.as_context_mut().0))?
             .instantiate_async(store)
@@ -1161,7 +1193,10 @@ impl<T> Linker<T> {
     /// # Ok(())
     /// # }
     /// ```
-    pub fn instantiate_pre(&self, module: &Module) -> Result<InstancePre<T>> {
+    pub fn instantiate_pre(&self, module: &Module) -> Result<InstancePre<T>>
+    where
+        T: 'static,
+    {
         self._instantiate_pre(module, None)
     }
 
@@ -1180,7 +1215,10 @@ impl<T> Linker<T> {
         &self,
         module: &Module,
         store: Option<&StoreOpaque>,
-    ) -> Result<InstancePre<T>> {
+    ) -> Result<InstancePre<T>>
+    where
+        T: 'static,
+    {
         let mut imports = module
             .imports()
             .map(|import| self._get_by_import(&import))
@@ -1210,7 +1248,10 @@ impl<T> Linker<T> {
     pub fn iter<'a: 'p, 'p>(
         &'a self,
         mut store: impl AsContextMut<Data = T> + 'p,
-    ) -> impl Iterator<Item = (&'a str, &'a str, Extern)> + 'p {
+    ) -> impl Iterator<Item = (&'a str, &'a str, Extern)> + 'p
+    where
+        T: 'static,
+    {
         self.map.iter().map(move |(key, item)| {
             let store = store.as_context_mut();
             (
@@ -1237,7 +1278,10 @@ impl<T> Linker<T> {
         mut store: impl AsContextMut<Data = T>,
         module: &str,
         name: &str,
-    ) -> Option<Extern> {
+    ) -> Option<Extern>
+    where
+        T: 'static,
+    {
         let store = store.as_context_mut().0;
         // Should be safe since `T` is connecting the linker and store
         Some(unsafe { self._get(module, name)?.to_extern(store) })
@@ -1264,7 +1308,10 @@ impl<T> Linker<T> {
         &self,
         mut store: impl AsContextMut<Data = T>,
         import: &ImportType,
-    ) -> Option<Extern> {
+    ) -> Option<Extern>
+    where
+        T: 'static,
+    {
         let store = store.as_context_mut().0;
         // Should be safe since `T` is connecting the linker and store
         Some(unsafe { self._get_by_import(import).ok()?.to_extern(store) })
@@ -1287,16 +1334,15 @@ impl<T> Linker<T> {
     /// Panics if the default function found is not owned by `store`. This
     /// function will also panic if the `store` provided does not come from the
     /// same [`Engine`] that this linker was created with.
-    pub fn get_default(
-        &self,
-        mut store: impl AsContextMut<Data = T>,
-        module: &str,
-    ) -> Result<Func> {
+    pub fn get_default(&self, mut store: impl AsContextMut<Data = T>, module: &str) -> Result<Func>
+    where
+        T: 'static,
+    {
         if let Some(external) = self.get(&mut store, module, "") {
             if let Extern::Func(func) = external {
                 return Ok(func);
             }
-            bail!("default export in '{}' is not a function", module);
+            bail!("default export in '{module}' is not a function");
         }
 
         // For compatibility, also recognize "_start".
@@ -1304,7 +1350,7 @@ impl<T> Linker<T> {
             if let Extern::Func(func) = external {
                 return Ok(func);
             }
-            bail!("`_start` in '{}' is not a function", module);
+            bail!("`_start` in '{module}' is not a function");
         }
 
         // Otherwise return a no-op function.
@@ -1312,7 +1358,7 @@ impl<T> Linker<T> {
     }
 }
 
-impl<T> Default for Linker<T> {
+impl<T: 'static> Default for Linker<T> {
     fn default() -> Linker<T> {
         Linker::new(&Engine::default())
     }
@@ -1331,13 +1377,20 @@ impl Definition {
         }
     }
 
+    /// Inserts this definition into the `store` provided.
+    ///
+    /// # Safety
+    ///
     /// Note the unsafety here is due to calling `HostFunc::to_func`. The
     /// requirement here is that the `T` that was originally used to create the
     /// `HostFunc` matches the `T` on the store.
     pub(crate) unsafe fn to_extern(&self, store: &mut StoreOpaque) -> Extern {
         match self {
             Definition::Extern(e, _) => e.clone(),
-            Definition::HostFunc(func) => func.to_func(store).into(),
+            // SAFETY: the contract of this function is the same as what's
+            // required of `to_func`, that `T` of the store matches the `T` of
+            // this original definition.
+            Definition::HostFunc(func) => unsafe { func.to_func(store).into() },
         }
     }
 
@@ -1357,7 +1410,7 @@ impl Definition {
                 *size = m.size();
             }
             Definition::Extern(Extern::Table(m), DefinitionType::Table(_, size)) => {
-                *size = m.internal_size(store);
+                *size = m._size(store);
             }
             _ => {}
         }
@@ -1366,16 +1419,15 @@ impl Definition {
 
 impl DefinitionType {
     pub(crate) fn from(store: &StoreOpaque, item: &Extern) -> DefinitionType {
-        let data = store.store_data();
         match item {
-            Extern::Func(f) => DefinitionType::Func(f.type_index(data)),
-            Extern::Table(t) => DefinitionType::Table(*t.wasmtime_ty(data), t.internal_size(store)),
-            Extern::Global(t) => DefinitionType::Global(*t.wasmtime_ty(data)),
+            Extern::Func(f) => DefinitionType::Func(f.type_index(store)),
+            Extern::Table(t) => DefinitionType::Table(*t.wasmtime_ty(store), t._size(store)),
+            Extern::Global(t) => DefinitionType::Global(*t.wasmtime_ty(store)),
             Extern::Memory(t) => {
-                DefinitionType::Memory(*t.wasmtime_ty(data), t.internal_size(store))
+                DefinitionType::Memory(*t.wasmtime_ty(store), t.internal_size(store))
             }
             Extern::SharedMemory(t) => DefinitionType::Memory(*t.ty().wasmtime_memory(), t.size()),
-            Extern::Tag(t) => DefinitionType::Tag(*t.wasmtime_ty(data)),
+            Extern::Tag(t) => DefinitionType::Tag(*t.wasmtime_ty(store)),
         }
     }
 
