@@ -7,6 +7,7 @@
 //! cranelift-compiled adapters, will use this `VMComponentContext` as well.
 
 use crate::component::{Component, Instance, InstancePre, ResourceType, RuntimeImport};
+use crate::module::ModuleRegistry;
 use crate::runtime::component::ComponentInstanceId;
 use crate::runtime::vm::instance::{InstanceLayout, OwnedInstance, OwnedVMContext};
 use crate::runtime::vm::vmcontext::VMFunctionBody;
@@ -77,6 +78,12 @@ pub struct ComponentInstance {
     // Otherwise the full guts of this component should only ever be used during
     // the instantiation of this instance, meaning that after instantiation much
     // of the component can be thrown away (theoretically).
+    //
+    // SAFETY: this field cannot be overwritten after an instance is created. It
+    // must contain this exact same value for the entire lifetime of this
+    // instance. This enables borrowing the component and this instance at the
+    // same time (instance mutably, component not). Additionally it enables
+    // borrowing a store mutably at the same time as a contained instance.
     component: Component,
 
     /// State of handles (e.g. resources, waitables, etc.) for this component.
@@ -337,11 +344,11 @@ impl ComponentInstance {
     ///
     /// This can only be called after `idx` has been initialized at runtime
     /// during the instantiation process of a component.
-    pub fn runtime_memory(&self, idx: RuntimeMemoryIndex) -> *mut VMMemoryDefinition {
+    pub fn runtime_memory(&self, idx: RuntimeMemoryIndex) -> NonNull<VMMemoryDefinition> {
         unsafe {
             let ret = *self.vmctx_plus_offset::<VmPtr<_>>(self.offsets.runtime_memory(idx));
             debug_assert!(ret.as_ptr() as usize != INVALID_PTR);
-            ret.as_ptr()
+            ret.as_non_null()
         }
     }
 
@@ -362,6 +369,7 @@ impl ComponentInstance {
     /// Returns the `Func` at index `func_idx` in the funcref table at `table_idx`.
     pub fn index_runtime_func_table(
         &self,
+        registry: &ModuleRegistry,
         table_idx: RuntimeTableIndex,
         func_idx: u64,
     ) -> Result<Option<Func>> {
@@ -375,7 +383,8 @@ impl ComponentInstance {
             // SAFETY: We just constructed `instance_ptr` from a valid pointer. This pointer won't leave
             // this call, so we don't need a lifetime to bind it to.
             let instance = Pin::new_unchecked(instance_ptr.as_mut());
-            let table = instance.get_defined_table_with_lazy_init(table.index, [func_idx]);
+            let table =
+                instance.get_defined_table_with_lazy_init(registry, table.index, [func_idx]);
             let func = table
                 .get_func(func_idx)?
                 .map(|funcref| Func::from_vm_func_ref(store.id(), funcref));
