@@ -19,7 +19,8 @@ use std::time::Duration;
 const CRATES_TO_PUBLISH: &[&str] = &[
     // pulley
     "cranelift-bitset",
-    "wasmtime-math",
+    "wasmtime-internal-math",
+    "pulley-macros",
     "pulley-interpreter",
     // cranelift
     "cranelift-srcgen",
@@ -40,7 +41,10 @@ const CRATES_TO_PUBLISH: &[&str] = &[
     "cranelift-native",
     "cranelift-object",
     "cranelift-interpreter",
-    "wasmtime-jit-icache-coherence",
+    "wasmtime-internal-jit-icache-coherence",
+    // Wasmtime unwinder, used by both `cranelift-jit` (optionally) and filetests, and by Wasmtime.
+    "wasmtime-internal-unwinder",
+    // Cranelift crates that use Wasmtime unwinder.
     "cranelift-jit",
     "cranelift",
     // wiggle
@@ -49,20 +53,19 @@ const CRATES_TO_PUBLISH: &[&str] = &[
     // winch
     "winch",
     // wasmtime
-    "wasmtime-asm-macros",
-    "wasmtime-versioned-export-macros",
-    "wasmtime-slab",
-    "wasmtime-component-util",
-    "wasmtime-wit-bindgen",
-    "wasmtime-component-macro",
-    "wasmtime-jit-debug",
-    "wasmtime-fiber",
+    "wasmtime-internal-versioned-export-macros",
+    "wasmtime-internal-slab",
+    "wasmtime-internal-component-util",
+    "wasmtime-internal-wit-bindgen",
+    "wasmtime-internal-component-macro",
+    "wasmtime-internal-jit-debug",
+    "wasmtime-internal-fiber",
     "wasmtime-environ",
-    "wasmtime-wmemcheck",
-    "wasmtime-cranelift",
-    "wasmtime-cache",
+    "wasmtime-internal-wmemcheck",
+    "wasmtime-internal-cranelift",
+    "wasmtime-internal-cache",
     "winch-codegen",
-    "wasmtime-winch",
+    "wasmtime-internal-winch",
     "wasmtime",
     // wasi-common/wiggle
     "wiggle",
@@ -76,11 +79,13 @@ const CRATES_TO_PUBLISH: &[&str] = &[
     "wasmtime-wasi-keyvalue",
     "wasmtime-wasi-threads",
     "wasmtime-wasi-tls",
+    "wasmtime-wasi-tls-nativetls",
     "wasmtime-wast",
-    "wasmtime-c-api-macros",
+    "wasmtime-internal-c-api-macros",
     "wasmtime-c-api-impl",
+    "wasmtime-wizer",
     "wasmtime-cli-flags",
-    "wasmtime-explorer",
+    "wasmtime-internal-explorer",
     "wasmtime-cli",
 ];
 
@@ -95,12 +100,14 @@ const PUBLIC_CRATES: &[&str] = &[
     "wasmtime-wasi-io",
     "wasmtime-wasi",
     "wasmtime-wasi-tls",
+    "wasmtime-wasi-tls-nativetls",
     "wasmtime-wasi-http",
     "wasmtime-wasi-nn",
     "wasmtime-wasi-config",
     "wasmtime-wasi-keyvalue",
     "wasmtime-wasi-threads",
     "wasmtime-cli",
+    "wasmtime-wizer",
     // All cranelift crates are considered "public" in that they can't have
     // breaking API changes in patch releases.
     "cranelift-srcgen",
@@ -336,7 +343,10 @@ fn bump_version(krate: &Crate, crates: &[Crate], patch: bool) {
             if !other.publish {
                 continue;
             }
-            if !is_deps || !line.starts_with(&format!("{} ", other.name)) {
+            if !is_deps
+                || (!line.starts_with(&format!("{} ", other.name))
+                    && !line.contains(&format!("package = '{}'", other.name)))
+            {
                 continue;
             }
             if !line.contains(&other.version) {
@@ -437,14 +447,13 @@ fn publish(krate: &Crate) -> bool {
 
     // First make sure the crate isn't already published at this version. This
     // script may be re-run and there's no need to re-attempt previous work.
-    let output = cmd_output(Command::new("curl").arg(&format!(
+    let Some(output) = curl(&format!(
         "https://crates.io/api/v1/crates/{}/versions",
         krate.name
-    )));
-    if output.status.success()
-        && String::from_utf8_lossy(&output.stdout)
-            .contains(&format!("\"num\":\"{}\"", krate.version))
-    {
+    )) else {
+        return false;
+    };
+    if output.contains(&format!("\"num\":\"{}\"", krate.version)) {
         println!(
             "skip publish {} because {} is already published",
             krate.name, krate.version,
@@ -466,13 +475,13 @@ fn publish(krate: &Crate) -> bool {
     // After we've published then make sure that the `wasmtime-publish` group is
     // added to this crate for future publications. If it's already present
     // though we can skip the `cargo owner` modification.
-    let output = cmd_output(Command::new("curl").arg(&format!(
+    let Some(output) = curl(&format!(
         "https://crates.io/api/v1/crates/{}/owners",
         krate.name
-    )));
-    if output.status.success()
-        && String::from_utf8_lossy(&output.stdout).contains("wasmtime-publish")
-    {
+    )) else {
+        return false;
+    };
+    if output.contains("wasmtime-publish") {
         println!(
             "wasmtime-publish already listed as an owner of {}",
             krate.name
@@ -494,6 +503,21 @@ fn publish(krate: &Crate) -> bool {
     true
 }
 
+fn curl(url: &str) -> Option<String> {
+    let output = cmd_output(
+        Command::new("curl")
+            .arg("--user-agent")
+            .arg("bytecodealliance/wasmtime auto-publish script")
+            .arg(url),
+    );
+    if !output.status.success() {
+        println!("failed to curl: {}", output.status);
+        println!("stderr: {}", String::from_utf8_lossy(&output.stderr));
+        return None;
+    }
+    Some(String::from_utf8_lossy(&output.stdout).into())
+}
+
 // Verify the current tree is publish-able to crates.io. The intention here is
 // that we'll run `cargo package` on everything which verifies the build as-if
 // it were published to crates.io. This requires using an incrementally-built
@@ -527,6 +551,8 @@ fn verify(crates: &[Crate]) {
     }
 
     fn verify_and_vendor(krate: &Crate) {
+        verify_crates_io(krate);
+
         let mut cmd = Command::new("cargo");
         cmd.arg("package")
             .arg("--manifest-path")
@@ -584,4 +610,30 @@ fn verify(crates: &[Crate]) {
             version
         );
     }
+
+    fn verify_crates_io(krate: &Crate) {
+        let name = &krate.name;
+        let Some(owners) = curl(&format!("https://crates.io/api/v1/crates/{name}/owners")) else {
+            panic!("failed to get owners for {name}", name = name);
+        };
+
+        let assert_owner = |owner: &str| {
+            let owner_json = format!("\"{owner}\"");
+            if !owners.contains(&owner_json) {
+                panic!(
+                    "
+crate {name} is not owned by {owner}, please run:
+
+    cargo owner -a {owner} {name}
+",
+                    name = name
+                );
+            }
+        };
+
+        // the wasmtime-publish github user
+        assert_owner("wasmtime-publish");
+        // the BA team which can publish crates
+        assert_owner("github:bytecodealliance:wasmtime-publish");
+    }
 }
