@@ -1,5 +1,7 @@
 use crate::async_functions::{PollOnce, execute_across_threads};
 use anyhow::Result;
+use std::pin::Pin;
+use std::task::{Context, Poll};
 use wasmtime::{AsContextMut, Config, component::*};
 use wasmtime::{Engine, Store, StoreContextMut, Trap};
 use wasmtime_component_util::REALLOC_AND_FREE;
@@ -342,7 +344,7 @@ async fn task_deletion() -> Result<()> {
             (import "" "mem" (memory 1))
             (import "" "task.return" (func $task-return (param i32)))
             (import "" "task.cancel" (func $task-cancel))
-            (import "" "thread.new_indirect" (func $thread-new-indirect (param i32 i32) (result i32)))
+            (import "" "thread.new-indirect" (func $thread-new-indirect (param i32 i32) (result i32)))
             (import "" "thread.suspend" (func $thread-suspend (result i32)))
             (import "" "thread.suspend-cancellable" (func $thread-suspend-cancellable (result i32)))
             (import "" "thread.yield-to" (func $thread-yield-to (param i32) (result i32)))
@@ -370,8 +372,8 @@ async fn task_deletion() -> Result<()> {
                 (drop (call $thread-suspend)))
 
             (func $yield-loop (param i32)
-                (loop $top 
-                    (drop (call $thread-yield)) 
+                (loop $top
+                    (drop (call $thread-yield))
                     (br $top)))
 
             (func (export "explicit-thread-calls-return-stackful")
@@ -385,7 +387,7 @@ async fn task_deletion() -> Result<()> {
 
             (func (export "cb") (param i32 i32 i32) (result i32)
                 (unreachable))
-            
+
             (func (export "explicit-thread-suspends-sync") (result i32)
                 (call $thread-resume-later
                     (call $thread-new-indirect (global.get $suspend-ftbl-idx) (i32.const 42)))
@@ -417,24 +419,23 @@ async fn task_deletion() -> Result<()> {
                     (call $thread-new-indirect (global.get $suspend-ftbl-idx) (i32.const 42)))
                 (call $task-return (i32.const 42))
                 (i32.const 0 (; EXIT ;)))
-            
 
-            ;; Initialize the function table that will be used by thread.new_indirect
+            ;; Initialize the function table that will be used by thread.new-indirect
             (elem (table $indirect-function-table) (i32.const 0 (; call-return-ftbl-idx ;)) func $call-return)
             (elem (table $indirect-function-table) (i32.const 1 (; suspend-ftbl-idx ;)) func $suspend)
             (elem (table $indirect-function-table) (i32.const 2 (; yield-loop-ftbl-idx ;)) func $yield-loop)
-        ) 
+        )
 
         ;; Instantiate the libc module to get the table
         (core instance $libc (instantiate $libc))
-        ;; Get access to `thread.new_indirect` that uses the table from libc
+        ;; Get access to `thread.new-indirect` that uses the table from libc
         (core type $start-func-ty (func (param i32)))
         (alias core export $libc "__indirect_function_table" (core table $indirect-function-table))
 
         (core func $task-return (canon task.return (result u32)))
         (core func $task-cancel (canon task.cancel))
-        (core func $thread-new-indirect 
-            (canon thread.new_indirect $start-func-ty (table $indirect-function-table)))
+        (core func $thread-new-indirect
+            (canon thread.new-indirect $start-func-ty (table $indirect-function-table)))
         (core func $thread-yield (canon thread.yield))
         (core func $thread-yield-cancellable (canon thread.yield cancellable))
         (core func $thread-index (canon thread.index))
@@ -456,7 +457,7 @@ async fn task_deletion() -> Result<()> {
                     (export "mem" (memory $memory "mem"))
                     (export "task.return" (func $task-return))
                     (export "task.cancel" (func $task-cancel))
-                    (export "thread.new_indirect" (func $thread-new-indirect))
+                    (export "thread.new-indirect" (func $thread-new-indirect))
                     (export "thread.index" (func $thread-index))
                     (export "thread.yield-to" (func $thread-yield-to))
                     (export "thread.yield-to-cancellable" (func $thread-yield-to-cancellable))
@@ -472,33 +473,33 @@ async fn task_deletion() -> Result<()> {
                     (export "waitable-set.new" (func $waitable-set.new))))
                 (with "libc" (instance $libc))))
 
-        (func (export "explicit-thread-calls-return-stackful") (result u32) 
+        (func (export "explicit-thread-calls-return-stackful") async (result u32)
             (canon lift (core func $cm "explicit-thread-calls-return-stackful") async))
-        (func (export "explicit-thread-calls-return-stackless") (result u32) 
+        (func (export "explicit-thread-calls-return-stackless") async (result u32)
             (canon lift (core func $cm "explicit-thread-calls-return-stackless") async (callback (func $cm "cb"))))
-        (func (export "explicit-thread-suspends-sync") (result u32) 
+        (func (export "explicit-thread-suspends-sync") async (result u32)
             (canon lift (core func $cm "explicit-thread-suspends-sync")))
-        (func (export "explicit-thread-suspends-stackful") (result u32) 
+        (func (export "explicit-thread-suspends-stackful") async (result u32)
             (canon lift (core func $cm "explicit-thread-suspends-stackful") async))
-        (func (export "explicit-thread-suspends-stackless") (result u32) 
+        (func (export "explicit-thread-suspends-stackless") async (result u32)
             (canon lift (core func $cm "explicit-thread-suspends-stackless") async (callback (func $cm "cb"))))
-        (func (export "explicit-thread-yield-loops-sync") (result u32) 
+        (func (export "explicit-thread-yield-loops-sync") async (result u32)
             (canon lift (core func $cm "explicit-thread-yield-loops-sync")))
-        (func (export "explicit-thread-yield-loops-stackful") (result u32) 
+        (func (export "explicit-thread-yield-loops-stackful") async (result u32)
             (canon lift (core func $cm "explicit-thread-yield-loops-stackful") async))
-        (func (export "explicit-thread-yield-loops-stackless") (result u32) 
+        (func (export "explicit-thread-yield-loops-stackless") async (result u32)
             (canon lift (core func $cm "explicit-thread-yield-loops-stackless") async (callback (func $cm "cb"))))
     )
 
-    (component $D 
-        (import "explicit-thread-calls-return-stackful" (func $explicit-thread-calls-return-stackful (result u32)))
-        (import "explicit-thread-calls-return-stackless" (func $explicit-thread-calls-return-stackless (result u32)))
-        (import "explicit-thread-suspends-sync" (func $explicit-thread-suspends-sync (result u32)))
-        (import "explicit-thread-suspends-stackful" (func $explicit-thread-suspends-stackful (result u32)))
-        (import "explicit-thread-suspends-stackless" (func $explicit-thread-suspends-stackless (result u32)))
-        (import "explicit-thread-yield-loops-sync" (func $explicit-thread-yield-loops-sync (result u32)))
-        (import "explicit-thread-yield-loops-stackful" (func $explicit-thread-yield-loops-stackful (result u32)))
-        (import "explicit-thread-yield-loops-stackless" (func $explicit-thread-yield-loops-stackless (result u32)))
+    (component $D
+        (import "explicit-thread-calls-return-stackful" (func $explicit-thread-calls-return-stackful async (result u32)))
+        (import "explicit-thread-calls-return-stackless" (func $explicit-thread-calls-return-stackless async (result u32)))
+        (import "explicit-thread-suspends-sync" (func $explicit-thread-suspends-sync async (result u32)))
+        (import "explicit-thread-suspends-stackful" (func $explicit-thread-suspends-stackful async (result u32)))
+        (import "explicit-thread-suspends-stackless" (func $explicit-thread-suspends-stackless async (result u32)))
+        (import "explicit-thread-yield-loops-sync" (func $explicit-thread-yield-loops-sync async (result u32)))
+        (import "explicit-thread-yield-loops-stackful" (func $explicit-thread-yield-loops-stackful async (result u32)))
+        (import "explicit-thread-yield-loops-stackless" (func $explicit-thread-yield-loops-stackless async (result u32)))
 
         (core module $Memory (memory (export "mem") 1))
         (core instance $memory (instantiate $Memory))
@@ -541,7 +542,7 @@ async fn task_deletion() -> Result<()> {
 
                 (if (i32.eq (i32.and (local.get 0) (i32.const 0xF)) (i32.const 2 (; RETURNED ;)))
                     (then (call $check (i32.load (local.get $retp))))
-                    (else 
+                    (else
                         (call $waitable.join (i32.shr_u (local.get 0) (i32.const 4)) (local.get $ws))
                         (drop (call $waitable-set.wait (local.get $ws) (local.get $ws-retp)))
                         (call $check (i32.load (local.get $retp)))))
@@ -558,7 +559,7 @@ async fn task_deletion() -> Result<()> {
                 (call $check (call $explicit-thread-yield-loops-sync))
                 (call $check (call $explicit-thread-yield-loops-stackful))
                 (call $check (call $explicit-thread-yield-loops-stackless))
-                
+
                 (call $check-async (call $explicit-thread-calls-return-stackless-async (local.get $retp)))
                 (call $check-async (call $explicit-thread-calls-return-stackful-async (local.get $retp)))
                 (call $check-async (call $explicit-thread-suspends-sync-async (local.get $retp)))
@@ -619,7 +620,7 @@ async fn task_deletion() -> Result<()> {
             (export "subtask.cancel" (func $subtask.cancel))
             (export "thread.yield" (func $thread.yield))
         ))))
-        (func (export "run") (result u32) (canon lift (core func $dm "run")))
+        (func (export "run") async (result u32) (canon lift (core func $dm "run")))
     )
 
     (instance $c (instantiate $C))
@@ -671,3 +672,98 @@ async fn task_deletion() -> Result<()> {
 
     Ok(())
 }
+
+#[tokio::test]
+#[cfg_attr(miri, ignore)]
+async fn cancel_host_future() -> Result<()> {
+    let mut config = Config::new();
+    config.async_support(true);
+    config.wasm_component_model_async(true);
+    let engine = Engine::new(&config)?;
+
+    let component = Component::new(
+        &engine,
+        r#"
+(component
+  (core module $libc (memory (export "memory") 1))
+  (core instance $libc (instantiate $libc))
+  (core module $m
+    (import "" "future.read" (func $future.read (param i32 i32) (result i32)))
+    (import "" "future.cancel-read" (func $future.cancel-read (param i32) (result i32)))
+    (memory (export "memory") 1)
+
+    (func (export "run") (param i32)
+      ;; read/cancel attempt 1
+      (call $future.read (local.get 0) (i32.const 100))
+      i32.const -1 ;; BLOCKED
+      i32.ne
+      if unreachable end
+
+      (call $future.cancel-read (local.get 0))
+      i32.const 2 ;; CANCELLED
+      i32.ne
+      if unreachable end
+
+      ;; read/cancel attempt 2
+      (call $future.read (local.get 0) (i32.const 100))
+      i32.const -1 ;; BLOCKED
+      i32.ne
+      if unreachable end
+
+      (call $future.cancel-read (local.get 0))
+      i32.const 2 ;; CANCELLED
+      i32.ne
+      if unreachable end
+    )
+  )
+
+  (type $f (future u32))
+  (core func $future.read (canon future.read $f async (memory $libc "memory")))
+  (core func $future.cancel-read (canon future.cancel-read $f))
+
+  (core instance $i (instantiate $m
+    (with "" (instance
+      (export "future.read" (func $future.read))
+      (export "future.cancel-read" (func $future.cancel-read))
+    ))
+  ))
+
+  (func (export "run") async (param "f" $f)
+    (canon lift
+      (core func $i "run")
+      (memory $libc "memory")
+    )
+  )
+)
+        "#,
+    )?;
+
+    let mut store = Store::new(&engine, ());
+    let instance = Linker::new(&engine)
+        .instantiate_async(&mut store, &component)
+        .await?;
+    let func = instance.get_typed_func::<(FutureReader<u32>,), ()>(&mut store, "run")?;
+    let reader = FutureReader::new(&mut store, MyFutureReader);
+    func.call_async(&mut store, (reader,)).await?;
+
+    return Ok(());
+
+    struct MyFutureReader;
+
+    impl FutureProducer<()> for MyFutureReader {
+        type Item = u32;
+
+        fn poll_produce(
+            self: Pin<&mut Self>,
+            _cx: &mut Context<'_>,
+            _store: StoreContextMut<()>,
+            finish: bool,
+        ) -> Poll<Result<Option<Self::Item>>> {
+            if finish {
+                Poll::Ready(Ok(None))
+            } else {
+                Poll::Pending
+            }
+        }
+    }
+}
