@@ -1,18 +1,18 @@
-use anyhow::{bail, ensure, Result};
+use anyhow::{Result, bail, ensure};
 use wasmparser::{Ieee32, Ieee64};
 use wasmtime_environ::{VMOffsets, WasmHeapType, WasmValType};
 
 use super::ControlStackFrame;
 use crate::{
-    abi::{scratch, vmctx, ABIOperand, ABIResults, RetArea},
-    codegen::{CodeGenError, CodeGenPhase, Emission, Prologue},
+    abi::{ABIOperand, ABIResults, RetArea, vmctx},
+    codegen::{BranchState, CodeGenError, CodeGenPhase, Emission, Prologue},
     frame::Frame,
     isa::reg::RegClass,
     masm::{
-        ExtractLaneKind, MacroAssembler, OperandSize, RegImm, ReplaceLaneKind, SPOffset, ShiftKind,
-        StackSlot,
+        ExtractLaneKind, Imm, IntScratch, MacroAssembler, MemMoveDirection, OperandSize, RegImm,
+        ReplaceLaneKind, SPOffset, ShiftKind, StackSlot,
     },
-    reg::{writable, Reg, WritableReg},
+    reg::{Reg, WritableReg, writable},
     regalloc::RegAlloc,
     stack::{Stack, TypedReg, Val},
 };
@@ -64,7 +64,7 @@ impl<'a> CodeGenContext<'a, Emission> {
             let typed_reg = self.pop_to_reg(masm, None)?;
             masm.shift_ir(
                 writable!(typed_reg.reg),
-                val as u64,
+                Imm::i32(val),
                 typed_reg.reg,
                 kind,
                 OperandSize::S32,
@@ -93,7 +93,7 @@ impl<'a> CodeGenContext<'a, Emission> {
             let typed_reg = self.pop_to_reg(masm, None)?;
             masm.shift_ir(
                 writable!(typed_reg.reg),
-                val as u64,
+                Imm::i64(val),
                 typed_reg.reg,
                 kind,
                 OperandSize::S64,
@@ -288,15 +288,17 @@ impl<'a> CodeGenContext<'a, Emission> {
             Val::V128(v) => masm.store(RegImm::v128(v), addr, size)?,
             Val::Local(local) => {
                 let slot = self.frame.get_wasm_local(local.index);
-                let scratch = scratch!(M);
                 let local_addr = masm.local_address(&slot)?;
-                masm.load(local_addr, writable!(scratch), size)?;
-                masm.store(scratch.into(), addr, size)?;
+                masm.with_scratch::<IntScratch, _>(|masm, scratch| {
+                    masm.load(local_addr, scratch.writable(), size)?;
+                    masm.store(scratch.inner().into(), addr, size)
+                })?;
             }
             Val::Memory(_) => {
-                let scratch = scratch!(M, &ty);
-                masm.pop(writable!(scratch), size)?;
-                masm.store(scratch.into(), addr, size)?;
+                masm.with_scratch_for(ty, |masm, scratch| {
+                    masm.pop(scratch.writable(), size)?;
+                    masm.store(scratch.inner().into(), addr, size)
+                })?;
             }
         }
 
@@ -355,7 +357,7 @@ impl<'a> CodeGenContext<'a, Emission> {
     {
         let src = self.pop_to_reg(masm, None)?;
         let dst = self.pop_to_reg(masm, None)?;
-        let dst = emit(masm, dst.reg, src.reg.into(), size)?;
+        let dst = emit(masm, dst.reg, src.reg, size)?;
         self.free_reg(src);
         self.stack.push(dst.into());
 
@@ -653,68 +655,105 @@ impl<'a> CodeGenContext<'a, Emission> {
         Self::spill_impl(&mut self.stack, &mut self.regalloc, &self.frame, masm)
     }
 
-    /// Prepares the compiler to emit an uncoditional jump to the given
-    /// destination branch.  This process involves:
-    /// * Balancing the machine
-    ///   stack pointer and value stack by popping it to match the destination
-    ///   branch.
+    /// Prepares the compiler to branch to the given destination
+    /// frame.
+    ///  This process involves:
+    /// * Balancing the machine stack pointer and value stack by
+    ///   popping it to match the destination branch.
     /// * Updating the reachability state.
     /// * Marking the destination frame as a destination target.
-    pub fn unconditional_jump<M, F>(
+    pub fn br<M, F, B>(
         &mut self,
         dest: &mut ControlStackFrame,
         masm: &mut M,
-        mut f: F,
+        mut maybe_pop_results: F,
     ) -> Result<()>
     where
         M: MacroAssembler,
         F: FnMut(&mut M, &mut Self, &mut ControlStackFrame) -> Result<()>,
+        B: BranchState,
     {
         let state = dest.stack_state();
         let target_offset = state.target_offset;
         let base_offset = state.base_offset;
-        // Invariant: The SP, must be greater or equal to the target
-        // SP, given that we haven't popped any results by this point
-        // yet. But it may happen in the callback.
+        let results_size = dest.results::<M>()?.size();
+
+        maybe_pop_results(masm, self, dest)?;
+        // After calling `maybe_pop_results`, the stack pointer plus
+        // any result space needed, must be greater or equal to the
+        // destination frame base stack pointer offset.
+        //
+        // We check
+        //   current_sp + results >= base_offset
+        // as opposed to
+        //   current_sp >= base_offset
+        //
+        // To:
+        //  - Verify that `maybe_pop_results` popped exactly the right
+        //    amount relative to the base offset.
+        //  - Accommodate for multi-branch cases (i.e., `br_table`) in which
+        //    result handling happens only once and _could_ happen outside of
+        //    `maybe_pop_results` callback.
+        //
+        //
+        // Ensuring that the current stack pointer offset plus any
+        // result space is equal to or greater than the target branch
+        // base offset is the the most deterministic check at branch
+        // emission time since we can be certain that the base offset
+        // is the value recorded when a new control frame was pushed,
+        // upon which the expected target offset is calculated.
         ensure!(
-            masm.sp_offset()?.as_u32() >= base_offset.as_u32(),
+            (masm.sp_offset()?.as_u32() + results_size) >= base_offset.as_u32(),
             CodeGenError::invalid_sp_offset()
         );
-        f(masm, self, dest)?;
 
-        // The following snippet, pops the stack pointer to ensure that it
-        // is correctly placed according to the expectations of the destination
-        // branch.
+        // At jump sites, the machine stack might be left unbalanced,
+        // due to register spills.
+        // The following snippet, pops the stack pointer to ensure
+        // that it is correctly placed according to the expectations
+        // of the destination branch.
         //
-        // This is done in the context of unconditional jumps, as the machine
-        // stack might be left unbalanced at the jump site, due to register
-        // spills. Note that in some cases the stack pointer offset might be
-        // already less than or equal to the original stack pointer offset
-        // registered when entering the destination control stack frame, which
-        // effectively means that when reaching the jump site no extra space was
-        // allocated similar to what would happen in a fall through in which we
-        // assume that the program has allocated and deallocated the right
-        // amount of stack space.
+        // Note that in most branch cases (`return`, ` br`) the stack
+        // pointer will be already balanced, by virtue of calling
+        // [`ControlStackFrame::pop_abi_results`] through the
+        // callback.
         //
-        // More generally speaking the current stack pointer will be less than
-        // the original stack pointer offset in cases in which the top value in
-        // the value stack is a memory entry which needs to be popped into the
-        // return location according to the ABI (a register for single value
-        // returns and a memory slot for 1+ returns). This could happen in the
-        // callback invocation above if the callback invokes
-        // `ControlStackFrame::pop_abi_results` (e.g. `br` instruction).
+        // More generally speaking the current stack pointer will be
+        // less than the destination frame stack pointer offset in
+        // cases in which the top value in the value stack is a memory
+        // entry which needs to be popped into the return location
+        // according to the ABI (a register for single value returns
+        // and a memory slot for 1+ returns).
         //
-        // After an unconditional jump, the compiler will enter in an
+        // Stack balancing is mostly required for WebAssembly
+        // instructions that deal with multiple destination branches
+        // (e.g., `br_table`) or fall-through scenarios (e.g.,
+        // `br_if`). In order to ensure that multi-value returns are
+        // handled correctly we ensure that correct placing of stack
+        // results by emitting a [`MacroAssembler::memmove`]
+        // instruction, prior to claiming any excess stack space.
+        //
+        // Depending on the branch state, the compiler might enter in an
         // unreachable state; instead of immediately truncating the value stack
         // to the expected length of the destination branch, we let the
         // reachability analysis code decide what should happen with the length
         // of the value stack once reachability is actually restored. At that
         // point, the right stack pointer offset will also be restored, which
         // should match the contents of the value stack.
+        if dest.unbalanced::<M>(masm)? {
+            masm.memmove(
+                masm.sp_offset()?,
+                target_offset,
+                results_size,
+                MemMoveDirection::LowToHigh,
+            )?;
+        }
         masm.ensure_sp_for_jump(target_offset)?;
         dest.set_as_target();
         masm.jmp(*dest.label())?;
-        self.reachable = false;
+        if B::unreachable_state_after_emission() {
+            self.reachable = false;
+        }
         Ok(())
     }
 
@@ -807,10 +846,12 @@ impl<'a> CodeGenContext<'a, Emission> {
                 Val::Local(local) => {
                     let slot = frame.get_wasm_local(local.index);
                     let addr = masm.local_address(&slot)?;
-                    let scratch = scratch!(M, &slot.ty);
-                    masm.load(addr, writable!(scratch), slot.ty.try_into()?)?;
-                    let stack_slot = masm.push(scratch, slot.ty.try_into()?)?;
-                    *v = Val::mem(slot.ty, stack_slot);
+                    masm.with_scratch_for(slot.ty, |masm, scratch| {
+                        masm.load(addr, scratch.writable(), slot.ty.try_into()?)?;
+                        let stack_slot = masm.push(scratch.inner(), slot.ty.try_into()?)?;
+                        *v = Val::mem(slot.ty, stack_slot);
+                        anyhow::Ok(())
+                    })?;
                 }
                 _ => {}
             }
